# This is a generated file! Please edit source .ksy file and use kaitai-struct-compiler to rebuild

from pkg_resources import parse_version
import kaitaistruct
from kaitaistruct import KaitaiStruct, KaitaiStream, BytesIO
from enum import Enum
import collections


if parse_version(kaitaistruct.__version__) < parse_version('0.9'):
    raise Exception("Incompatible Kaitai Struct Python API: 0.9 or later is required, but you have %s" % (kaitaistruct.__version__))

class PythonPyc27(KaitaiStruct):
    """Python interpreter runs .py files in 2 step process: first, it
    produces bytecode, which it then executes. Translation of .py source
    into bytecode is time-consuming, so Python dumps compiled bytecode
    into .pyc files, to be reused from cache at later time if possible.
    
    .pyc file is essentially a raw dump of `py_object` (see `body`) with
    a simple header prepended.
    """

    class Version(Enum):
        v15 = 20121
        v16 = 50428
        v20 = 50823
        v21 = 60202
        v22 = 60717
        v23_a0 = 62011
        v23_a0b = 62021
        v24_a0 = 62041
        v24_a3 = 62051
        v24_b1 = 62061
        v25_a0 = 62071
        v25_a0b = 62081
        v25_a0c = 62091
        v25_a0d = 62092
        v25_b3 = 62101
        v25_b3b = 62111
        v25_c1 = 62121
        v25_c2 = 62131
        v26_a0 = 62151
        v26_a1 = 62161
        v27_a0 = 62171
        v27_a0b = 62181
        v27_a0c = 62191
        v27_a0d = 62201
        v27_a0e = 62211
    SEQ_FIELDS = ["version_magic", "crlf", "modification_timestamp", "body"]
    def __init__(self, _io, _parent=None, _root=None):
        self._io = _io
        self._parent = _parent
        self._root = _root if _root else self
        self._debug = collections.defaultdict(dict)

    def _read(self):
        self._debug['version_magic']['start'] = self._io.pos()
        self.version_magic = KaitaiStream.resolve_enum(PythonPyc27.Version, self._io.read_u2le())
        self._debug['version_magic']['end'] = self._io.pos()
        self._debug['crlf']['start'] = self._io.pos()
        self.crlf = self._io.read_u2le()
        self._debug['crlf']['end'] = self._io.pos()
        self._debug['modification_timestamp']['start'] = self._io.pos()
        self.modification_timestamp = self._io.read_u4le()
        self._debug['modification_timestamp']['end'] = self._io.pos()
        self._debug['body']['start'] = self._io.pos()
        self.body = PythonPyc27.PyObject(self._io, self, self._root)
        self.body._read()
        self._debug['body']['end'] = self._io.pos()

    class CodeObject(KaitaiStruct):

        class FlagsEnum(Enum):
            has_args = 4
            has_kwargs = 8
            generator = 32
        SEQ_FIELDS = ["arg_count", "local_count", "stack_size", "flags", "code", "consts", "names", "var_names", "free_vars", "cell_vars", "filename", "name", "first_line_no", "lnotab"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['arg_count']['start'] = self._io.pos()
            self.arg_count = self._io.read_u4le()
            self._debug['arg_count']['end'] = self._io.pos()
            self._debug['local_count']['start'] = self._io.pos()
            self.local_count = self._io.read_u4le()
            self._debug['local_count']['end'] = self._io.pos()
            self._debug['stack_size']['start'] = self._io.pos()
            self.stack_size = self._io.read_u4le()
            self._debug['stack_size']['end'] = self._io.pos()
            self._debug['flags']['start'] = self._io.pos()
            self.flags = KaitaiStream.resolve_enum(PythonPyc27.CodeObject.FlagsEnum, self._io.read_u4le())
            self._debug['flags']['end'] = self._io.pos()
            self._debug['code']['start'] = self._io.pos()
            self.code = PythonPyc27.Assembly(self._io, self, self._root)
            self.code._read()
            self._debug['code']['end'] = self._io.pos()
            self._debug['consts']['start'] = self._io.pos()
            self.consts = PythonPyc27.PyObject(self._io, self, self._root)
            self.consts._read()
            self._debug['consts']['end'] = self._io.pos()
            self._debug['names']['start'] = self._io.pos()
            self.names = PythonPyc27.PyObject(self._io, self, self._root)
            self.names._read()
            self._debug['names']['end'] = self._io.pos()
            self._debug['var_names']['start'] = self._io.pos()
            self.var_names = PythonPyc27.PyObject(self._io, self, self._root)
            self.var_names._read()
            self._debug['var_names']['end'] = self._io.pos()
            self._debug['free_vars']['start'] = self._io.pos()
            self.free_vars = PythonPyc27.PyObject(self._io, self, self._root)
            self.free_vars._read()
            self._debug['free_vars']['end'] = self._io.pos()
            self._debug['cell_vars']['start'] = self._io.pos()
            self.cell_vars = PythonPyc27.PyObject(self._io, self, self._root)
            self.cell_vars._read()
            self._debug['cell_vars']['end'] = self._io.pos()
            self._debug['filename']['start'] = self._io.pos()
            self.filename = PythonPyc27.PyObject(self._io, self, self._root)
            self.filename._read()
            self._debug['filename']['end'] = self._io.pos()
            self._debug['name']['start'] = self._io.pos()
            self.name = PythonPyc27.PyObject(self._io, self, self._root)
            self.name._read()
            self._debug['name']['end'] = self._io.pos()
            self._debug['first_line_no']['start'] = self._io.pos()
            self.first_line_no = self._io.read_u4le()
            self._debug['first_line_no']['end'] = self._io.pos()
            self._debug['lnotab']['start'] = self._io.pos()
            self.lnotab = PythonPyc27.PyObject(self._io, self, self._root)
            self.lnotab._read()
            self._debug['lnotab']['end'] = self._io.pos()


    class Assembly(KaitaiStruct):
        SEQ_FIELDS = ["string_magic", "length", "items"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['string_magic']['start'] = self._io.pos()
            self.string_magic = self._io.read_bytes(1)
            self._debug['string_magic']['end'] = self._io.pos()
            if not self.string_magic == b"\x73":
                raise kaitaistruct.ValidationNotEqualError(b"\x73", self.string_magic, self._io, u"/types/assembly/seq/0")
            self._debug['length']['start'] = self._io.pos()
            self.length = self._io.read_u4le()
            self._debug['length']['end'] = self._io.pos()
            self._debug['items']['start'] = self._io.pos()
            self._raw_items = self._io.read_bytes(self.length)
            _io__raw_items = KaitaiStream(BytesIO(self._raw_items))
            self.items = PythonPyc27.OpArgs(_io__raw_items, self, self._root)
            self.items._read()
            self._debug['items']['end'] = self._io.pos()


    class OpArg(KaitaiStruct):

        class OpCodeEnum(Enum):
            stop_code = 0
            pop_top = 1
            rot_two = 2
            rot_three = 3
            dup_top = 4
            rot_four = 5
            nop = 9
            unary_positive = 10
            unary_negative = 11
            unary_not = 12
            unary_convert = 13
            unary_invert = 15
            binary_power = 19
            binary_multiply = 20
            binary_divide = 21
            binary_modulo = 22
            binary_add = 23
            binary_subtract = 24
            binary_subscr = 25
            binary_floor_divide = 26
            binary_true_divide = 27
            inplace_floor_divide = 28
            inplace_true_divide = 29
            slice_0 = 30
            slice_1 = 31
            slice_2 = 32
            slice_3 = 33
            store_slice_0 = 40
            store_slice_1 = 41
            store_slice_2 = 42
            store_slice_3 = 43
            delete_slice_0 = 50
            delete_slice_1 = 51
            delete_slice_2 = 52
            delete_slice_3 = 53
            store_map = 54
            inplace_add = 55
            inplace_subtract = 56
            inplace_multiply = 57
            inplace_divide = 58
            inplace_modulo = 59
            store_subscr = 60
            delete_subscr = 61
            binary_lshift = 62
            binary_rshift = 63
            binary_and = 64
            binary_xor = 65
            binary_or = 66
            inplace_power = 67
            get_iter = 68
            print_expr = 70
            print_item = 71
            print_newline = 72
            print_item_to = 73
            print_newline_to = 74
            inplace_lshift = 75
            inplace_rshift = 76
            inplace_and = 77
            inplace_xor = 78
            inplace_or = 79
            break_loop = 80
            with_cleanup = 81
            load_locals = 82
            return_value = 83
            import_star = 84
            exec_stmt = 85
            yield_value = 86
            pop_block = 87
            end_finally = 88
            build_class = 89
            store_name = 90
            delete_name = 91
            unpack_sequence = 92
            for_iter = 93
            list_append = 94
            store_attr = 95
            delete_attr = 96
            store_global = 97
            delete_global = 98
            dup_topx = 99
            load_const = 100
            load_name = 101
            build_tuple = 102
            build_list = 103
            build_set = 104
            build_map = 105
            load_attr = 106
            compare_op = 107
            import_name = 108
            import_from = 109
            jump_forward = 110
            jump_if_false_or_pop = 111
            jump_if_true_or_pop = 112
            jump_absolute = 113
            pop_jump_if_false = 114
            pop_jump_if_true = 115
            load_global = 116
            continue_loop = 119
            setup_loop = 120
            setup_except = 121
            setup_finally = 122
            load_fast = 124
            store_fast = 125
            delete_fast = 126
            raise_varargs = 130
            call_function = 131
            make_function = 132
            build_slice = 133
            make_closure = 134
            load_closure = 135
            load_deref = 136
            store_deref = 137
            call_function_var = 140
            call_function_kw = 141
            call_function_var_kw = 142
            setup_with = 143
            extended_arg = 145
            set_add = 146
            map_add = 147
        SEQ_FIELDS = ["op_code", "arg"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['op_code']['start'] = self._io.pos()
            self.op_code = KaitaiStream.resolve_enum(PythonPyc27.OpArg.OpCodeEnum, self._io.read_u1())
            self._debug['op_code']['end'] = self._io.pos()
            if self.op_code.value >= PythonPyc27.OpArg.OpCodeEnum.store_name.value:
                self._debug['arg']['start'] = self._io.pos()
                self.arg = self._io.read_u2le()
                self._debug['arg']['end'] = self._io.pos()



    class PyObject(KaitaiStruct):

        class ObjectType(Enum):
            tuple = 40
            py_false = 70
            none = 78
            string_ref = 82
            py_true = 84
            code_object = 99
            int = 105
            string = 115
            interned = 116
            unicode_string = 117
        SEQ_FIELDS = ["type", "value"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['type']['start'] = self._io.pos()
            self.type = KaitaiStream.resolve_enum(PythonPyc27.PyObject.ObjectType, self._io.read_u1())
            self._debug['type']['end'] = self._io.pos()
            self._debug['value']['start'] = self._io.pos()
            _on = self.type
            if _on == PythonPyc27.PyObject.ObjectType.string:
                self.value = PythonPyc27.PyObject.PyString(self._io, self, self._root)
                self.value._read()
            elif _on == PythonPyc27.PyObject.ObjectType.tuple:
                self.value = PythonPyc27.PyObject.Tuple(self._io, self, self._root)
                self.value._read()
            elif _on == PythonPyc27.PyObject.ObjectType.int:
                self.value = self._io.read_u4le()
            elif _on == PythonPyc27.PyObject.ObjectType.py_true:
                self.value = PythonPyc27.PyObject.PyTrue(self._io, self, self._root)
                self.value._read()
            elif _on == PythonPyc27.PyObject.ObjectType.py_false:
                self.value = PythonPyc27.PyObject.PyFalse(self._io, self, self._root)
                self.value._read()
            elif _on == PythonPyc27.PyObject.ObjectType.none:
                self.value = PythonPyc27.PyObject.PyNone(self._io, self, self._root)
                self.value._read()
            elif _on == PythonPyc27.PyObject.ObjectType.string_ref:
                self.value = PythonPyc27.PyObject.StringRef(self._io, self, self._root)
                self.value._read()
            elif _on == PythonPyc27.PyObject.ObjectType.code_object:
                self.value = PythonPyc27.CodeObject(self._io, self, self._root)
                self.value._read()
            elif _on == PythonPyc27.PyObject.ObjectType.interned:
                self.value = PythonPyc27.PyObject.InternedString(self._io, self, self._root)
                self.value._read()
            self._debug['value']['end'] = self._io.pos()

        class PyNone(KaitaiStruct):
            SEQ_FIELDS = []
            def __init__(self, _io, _parent=None, _root=None):
                self._io = _io
                self._parent = _parent
                self._root = _root if _root else self
                self._debug = collections.defaultdict(dict)

            def _read(self):
                pass


        class PyFalse(KaitaiStruct):
            SEQ_FIELDS = []
            def __init__(self, _io, _parent=None, _root=None):
                self._io = _io
                self._parent = _parent
                self._root = _root if _root else self
                self._debug = collections.defaultdict(dict)

            def _read(self):
                pass


        class StringRef(KaitaiStruct):
            SEQ_FIELDS = ["interned_list_index"]
            def __init__(self, _io, _parent=None, _root=None):
                self._io = _io
                self._parent = _parent
                self._root = _root if _root else self
                self._debug = collections.defaultdict(dict)

            def _read(self):
                self._debug['interned_list_index']['start'] = self._io.pos()
                self.interned_list_index = self._io.read_u4le()
                self._debug['interned_list_index']['end'] = self._io.pos()


        class PyTrue(KaitaiStruct):
            SEQ_FIELDS = []
            def __init__(self, _io, _parent=None, _root=None):
                self._io = _io
                self._parent = _parent
                self._root = _root if _root else self
                self._debug = collections.defaultdict(dict)

            def _read(self):
                pass


        class Tuple(KaitaiStruct):
            SEQ_FIELDS = ["count", "items"]
            def __init__(self, _io, _parent=None, _root=None):
                self._io = _io
                self._parent = _parent
                self._root = _root if _root else self
                self._debug = collections.defaultdict(dict)

            def _read(self):
                self._debug['count']['start'] = self._io.pos()
                self.count = self._io.read_u4le()
                self._debug['count']['end'] = self._io.pos()
                self._debug['items']['start'] = self._io.pos()
                self.items = [None] * (self.count)
                for i in range(self.count):
                    if not 'arr' in self._debug['items']:
                        self._debug['items']['arr'] = []
                    self._debug['items']['arr'].append({'start': self._io.pos()})
                    _t_items = PythonPyc27.PyObject(self._io, self, self._root)
                    _t_items._read()
                    self.items[i] = _t_items
                    self._debug['items']['arr'][i]['end'] = self._io.pos()

                self._debug['items']['end'] = self._io.pos()


        class UnicodeString(KaitaiStruct):
            SEQ_FIELDS = ["length", "data"]
            def __init__(self, _io, _parent=None, _root=None):
                self._io = _io
                self._parent = _parent
                self._root = _root if _root else self
                self._debug = collections.defaultdict(dict)

            def _read(self):
                self._debug['length']['start'] = self._io.pos()
                self.length = self._io.read_u4le()
                self._debug['length']['end'] = self._io.pos()
                self._debug['data']['start'] = self._io.pos()
                self.data = (self._io.read_bytes(self.length)).decode(u"utf-8")
                self._debug['data']['end'] = self._io.pos()


        class InternedString(KaitaiStruct):
            SEQ_FIELDS = ["length", "data"]
            def __init__(self, _io, _parent=None, _root=None):
                self._io = _io
                self._parent = _parent
                self._root = _root if _root else self
                self._debug = collections.defaultdict(dict)

            def _read(self):
                self._debug['length']['start'] = self._io.pos()
                self.length = self._io.read_u4le()
                self._debug['length']['end'] = self._io.pos()
                self._debug['data']['start'] = self._io.pos()
                self.data = (self._io.read_bytes(self.length)).decode(u"utf-8")
                self._debug['data']['end'] = self._io.pos()


        class PyString(KaitaiStruct):
            SEQ_FIELDS = ["length", "data"]
            def __init__(self, _io, _parent=None, _root=None):
                self._io = _io
                self._parent = _parent
                self._root = _root if _root else self
                self._debug = collections.defaultdict(dict)

            def _read(self):
                self._debug['length']['start'] = self._io.pos()
                self.length = self._io.read_u4le()
                self._debug['length']['end'] = self._io.pos()
                self._debug['data']['start'] = self._io.pos()
                self.data = self._io.read_bytes(self.length)
                self._debug['data']['end'] = self._io.pos()



    class OpArgs(KaitaiStruct):
        SEQ_FIELDS = ["items"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['items']['start'] = self._io.pos()
            self.items = []
            i = 0
            while not self._io.is_eof():
                if not 'arr' in self._debug['items']:
                    self._debug['items']['arr'] = []
                self._debug['items']['arr'].append({'start': self._io.pos()})
                _t_items = PythonPyc27.OpArg(self._io, self, self._root)
                _t_items._read()
                self.items.append(_t_items)
                self._debug['items']['arr'][len(self.items) - 1]['end'] = self._io.pos()
                i += 1

            self._debug['items']['end'] = self._io.pos()






import logging
import sys
from time import time
from typing import Collection, Iterator, Optional, TypeVar


STATUS = 15
TRACE = 5


class StatusLogHandler(logging.StreamHandler):
    def __init__(self, stream=sys.stderr.buffer):
        super().__init__(stream=stream)
        self._status: bytes = b''

    def print(self, text):
        if self._status and self.stream.isatty():
            self.stream.write(f"\r{' ' * len(self._status)}\r".encode('utf-8'))
            self.stream.flush()
            sys.stdout.write(str(text))
            sys.stdout.write("\n")
            sys.stdout.flush()
            self.stream.write(self._status)
            self.stream.flush()
        else:
            sys.stdout.write(str(text))
            sys.stdout.write("\n")

    def emit(self, record):
        isatty = self.stream.isatty()
        if self._status and isatty:
            self.stream.write(f"\r{' ' * len(self._status)}\r".encode('utf-8'))
        msg = record.getMessage()
        if isinstance(msg, str):
            msg = msg.encode('utf-8')
        if record.levelno == STATUS and isatty:
            self._status = msg
        else:
            self.stream.write(msg)
            self.stream.write(b'\n')
        if self._status and isatty:
            self.stream.write(self._status)
        self.stream.flush()


logging.addLevelName(STATUS, "STATUS")
logging.addLevelName(TRACE, "TRACE")

DEFAULT_STATUS_LOG_HANDLER = StatusLogHandler()


T = TypeVar("T")


class StatusLogger(logging.getLoggerClass()):
    def __init__(self, name, level=logging.NOTSET):
        super().__init__(name, level)
        self.addHandler(DEFAULT_STATUS_LOG_HANDLER)

    def status(self, msg, *args, **kwargs):
        if self.isEnabledFor(STATUS):
            self._log(STATUS, msg, args, **kwargs)

    def print(self, text: str):
        DEFAULT_STATUS_LOG_HANDLER.print(text)

    def range(
            self,
            iterable: Collection[T],
            desc: str = "",
            unit: str = "",
            delay: float = 0.0,
            update_interval: float = 0.5
    ) -> Iterator[T]:
        if not self.isEnabledFor(STATUS) or len(iterable) == 0:
            yield from iterable
            return
        size = len(iterable)
        start_time = time()
        last_update_time: float = 0.0
        last_percent: float = -1.0
        def print_msg(percent: float, item: int):
            msg = desc.strip()
            if msg:
                msg = f"{msg} "
            bar = "=" * int(30 * percent)
            if bar and percent < 1.0:
                bar = f"{bar[:-1]}>"
            bar = f"|{bar}{'-' * (30 - len(bar))}|"
            pct = f"{percent * 100.0:.1f}%"
            pct_pos = (len(bar) - len(pct)) // 2
            bar = f"{bar[:pct_pos]}{pct}{bar[pct_pos + len(pct):]}"
            msg = f"{msg}{bar} {item}/{size}{unit}"
            self.status(msg)
        if delay == 0.0:
            print_msg(0.0, 0)
        for i, obj in enumerate(iterable):
            current_time = time()
            elapsed_time = current_time - start_time
            if size == 1:
                new_percent = 1.0
            else:
                new_percent = i / (size - 1)
            if (elapsed_time >= delay and (
                    current_time - last_update_time >= update_interval or new_percent >= last_percent + 0.1
            )) or i == size - 1:
                last_update_time = current_time
                last_percent = new_percent
                print_msg(new_percent, i + 1)
            yield obj

    def trace(self, msg, *args, **kwargs):
        if self.isEnabledFor(TRACE):
            self._log(TRACE, msg, args, **kwargs)

    def clear_status(self):
        self.status("")


logging.setLoggerClass(StatusLogger)


def getStatusLogger(name) -> StatusLogger:
    return logging.getLogger(name)


def get_root_logger():
    l = getStatusLogger(__name__)
    while l.parent:
        l = l.parent
    return l


def setLevel(levelno):
    get_root_logger().setLevel(levelno)



# This is a generated file! Please edit source .ksy file and use kaitai-struct-compiler to rebuild

from pkg_resources import parse_version
import kaitaistruct
from kaitaistruct import KaitaiStruct, KaitaiStream, BytesIO
import collections


if parse_version(kaitaistruct.__version__) < parse_version('0.9'):
    raise Exception("Incompatible Kaitai Struct Python API: 0.9 or later is required, but you have %s" % (kaitaistruct.__version__))

class TcpSegment(KaitaiStruct):
    """TCP is one of the core Internet protocols on transport layer (AKA
    OSI layer 4), providing stateful connections with error checking,
    guarantees of delivery, order of segments and avoidance of duplicate
    delivery.
    """
    SEQ_FIELDS = ["src_port", "dst_port", "seq_num", "ack_num", "b12", "b13", "window_size", "checksum", "urgent_pointer", "body"]
    def __init__(self, _io, _parent=None, _root=None):
        self._io = _io
        self._parent = _parent
        self._root = _root if _root else self
        self._debug = collections.defaultdict(dict)

    def _read(self):
        self._debug['src_port']['start'] = self._io.pos()
        self.src_port = self._io.read_u2be()
        self._debug['src_port']['end'] = self._io.pos()
        self._debug['dst_port']['start'] = self._io.pos()
        self.dst_port = self._io.read_u2be()
        self._debug['dst_port']['end'] = self._io.pos()
        self._debug['seq_num']['start'] = self._io.pos()
        self.seq_num = self._io.read_u4be()
        self._debug['seq_num']['end'] = self._io.pos()
        self._debug['ack_num']['start'] = self._io.pos()
        self.ack_num = self._io.read_u4be()
        self._debug['ack_num']['end'] = self._io.pos()
        self._debug['b12']['start'] = self._io.pos()
        self.b12 = self._io.read_u1()
        self._debug['b12']['end'] = self._io.pos()
        self._debug['b13']['start'] = self._io.pos()
        self.b13 = self._io.read_u1()
        self._debug['b13']['end'] = self._io.pos()
        self._debug['window_size']['start'] = self._io.pos()
        self.window_size = self._io.read_u2be()
        self._debug['window_size']['end'] = self._io.pos()
        self._debug['checksum']['start'] = self._io.pos()
        self.checksum = self._io.read_u2be()
        self._debug['checksum']['end'] = self._io.pos()
        self._debug['urgent_pointer']['start'] = self._io.pos()
        self.urgent_pointer = self._io.read_u2be()
        self._debug['urgent_pointer']['end'] = self._io.pos()
        self._debug['body']['start'] = self._io.pos()
        self.body = self._io.read_bytes_full()
        self._debug['body']['end'] = self._io.pos()





# This is a generated file! Please edit source .ksy file and use kaitai-struct-compiler to rebuild

from pkg_resources import parse_version
import kaitaistruct
from kaitaistruct import KaitaiStruct, KaitaiStream, BytesIO
from enum import Enum
import collections


if parse_version(kaitaistruct.__version__) < parse_version('0.9'):
    raise Exception("Incompatible Kaitai Struct Python API: 0.9 or later is required, but you have %s" % (kaitaistruct.__version__))

from polyfile.kaitai.parsers import ethernet_frame
from polyfile.kaitai.parsers import windows_systemtime
class MicrosoftNetworkMonitorV2(KaitaiStruct):
    """Microsoft Network Monitor (AKA Netmon) is a proprietary Microsoft's
    network packet sniffing and analysis tool. It can save captured
    traffic as .cap files, which usually contain the packets and may
    contain some additional info - enhanced network info, calculated
    statistics, etc.
    
    There are at least 2 different versions of the format: v1 and
    v2. Netmon v3 seems to use the same file format as v1.
    
    .. seealso::
       Source - https://docs.microsoft.com/en-us/windows/win32/netmon2/capturefile-header-values
    """

    class Linktype(Enum):
        null_linktype = 0
        ethernet = 1
        ax25 = 3
        ieee802_5 = 6
        arcnet_bsd = 7
        slip = 8
        ppp = 9
        fddi = 10
        ppp_hdlc = 50
        ppp_ether = 51
        atm_rfc1483 = 100
        raw = 101
        c_hdlc = 104
        ieee802_11 = 105
        frelay = 107
        loop = 108
        linux_sll = 113
        ltalk = 114
        pflog = 117
        ieee802_11_prism = 119
        ip_over_fc = 122
        sunatm = 123
        ieee802_11_radiotap = 127
        arcnet_linux = 129
        apple_ip_over_ieee1394 = 138
        mtp2_with_phdr = 139
        mtp2 = 140
        mtp3 = 141
        sccp = 142
        docsis = 143
        linux_irda = 144
        user0 = 147
        user1 = 148
        user2 = 149
        user3 = 150
        user4 = 151
        user5 = 152
        user6 = 153
        user7 = 154
        user8 = 155
        user9 = 156
        user10 = 157
        user11 = 158
        user12 = 159
        user13 = 160
        user14 = 161
        user15 = 162
        ieee802_11_avs = 163
        bacnet_ms_tp = 165
        ppp_pppd = 166
        gprs_llc = 169
        gpf_t = 170
        gpf_f = 171
        linux_lapd = 177
        bluetooth_hci_h4 = 187
        usb_linux = 189
        ppi = 192
        ieee802_15_4 = 195
        sita = 196
        erf = 197
        bluetooth_hci_h4_with_phdr = 201
        ax25_kiss = 202
        lapd = 203
        ppp_with_dir = 204
        c_hdlc_with_dir = 205
        frelay_with_dir = 206
        ipmb_linux = 209
        ieee802_15_4_nonask_phy = 215
        usb_linux_mmapped = 220
        fc_2 = 224
        fc_2_with_frame_delims = 225
        ipnet = 226
        can_socketcan = 227
        ipv4 = 228
        ipv6 = 229
        ieee802_15_4_nofcs = 230
        dbus = 231
        dvb_ci = 235
        mux27010 = 236
        stanag_5066_d_pdu = 237
        nflog = 239
        netanalyzer = 240
        netanalyzer_transparent = 241
        ipoib = 242
        mpeg_2_ts = 243
        ng40 = 244
        nfc_llcp = 245
        infiniband = 247
        sctp = 248
        usbpcap = 249
        rtac_serial = 250
        bluetooth_le_ll = 251
        netlink = 253
        bluetooth_linux_monitor = 254
        bluetooth_bredr_bb = 255
        bluetooth_le_ll_with_phdr = 256
        profibus_dl = 257
        pktap = 258
        epon = 259
        ipmi_hpm_2 = 260
        zwave_r1_r2 = 261
        zwave_r3 = 262
        wattstopper_dlm = 263
        iso_14443 = 264
    SEQ_FIELDS = ["signature", "version_minor", "version_major", "mac_type", "time_capture_start", "frame_table_ofs", "frame_table_len", "user_data_ofs", "user_data_len", "comment_ofs", "comment_len", "statistics_ofs", "statistics_len", "network_info_ofs", "network_info_len", "conversation_stats_ofs", "conversation_stats_len"]
    def __init__(self, _io, _parent=None, _root=None):
        self._io = _io
        self._parent = _parent
        self._root = _root if _root else self
        self._debug = collections.defaultdict(dict)

    def _read(self):
        self._debug['signature']['start'] = self._io.pos()
        self.signature = self._io.read_bytes(4)
        self._debug['signature']['end'] = self._io.pos()
        if not self.signature == b"\x47\x4D\x42\x55":
            raise kaitaistruct.ValidationNotEqualError(b"\x47\x4D\x42\x55", self.signature, self._io, u"/seq/0")
        self._debug['version_minor']['start'] = self._io.pos()
        self.version_minor = self._io.read_u1()
        self._debug['version_minor']['end'] = self._io.pos()
        self._debug['version_major']['start'] = self._io.pos()
        self.version_major = self._io.read_u1()
        self._debug['version_major']['end'] = self._io.pos()
        self._debug['mac_type']['start'] = self._io.pos()
        self.mac_type = KaitaiStream.resolve_enum(MicrosoftNetworkMonitorV2.Linktype, self._io.read_u2le())
        self._debug['mac_type']['end'] = self._io.pos()
        self._debug['time_capture_start']['start'] = self._io.pos()
        self.time_capture_start = windows_systemtime.WindowsSystemtime(self._io)
        self.time_capture_start._read()
        self._debug['time_capture_start']['end'] = self._io.pos()
        self._debug['frame_table_ofs']['start'] = self._io.pos()
        self.frame_table_ofs = self._io.read_u4le()
        self._debug['frame_table_ofs']['end'] = self._io.pos()
        self._debug['frame_table_len']['start'] = self._io.pos()
        self.frame_table_len = self._io.read_u4le()
        self._debug['frame_table_len']['end'] = self._io.pos()
        self._debug['user_data_ofs']['start'] = self._io.pos()
        self.user_data_ofs = self._io.read_u4le()
        self._debug['user_data_ofs']['end'] = self._io.pos()
        self._debug['user_data_len']['start'] = self._io.pos()
        self.user_data_len = self._io.read_u4le()
        self._debug['user_data_len']['end'] = self._io.pos()
        self._debug['comment_ofs']['start'] = self._io.pos()
        self.comment_ofs = self._io.read_u4le()
        self._debug['comment_ofs']['end'] = self._io.pos()
        self._debug['comment_len']['start'] = self._io.pos()
        self.comment_len = self._io.read_u4le()
        self._debug['comment_len']['end'] = self._io.pos()
        self._debug['statistics_ofs']['start'] = self._io.pos()
        self.statistics_ofs = self._io.read_u4le()
        self._debug['statistics_ofs']['end'] = self._io.pos()
        self._debug['statistics_len']['start'] = self._io.pos()
        self.statistics_len = self._io.read_u4le()
        self._debug['statistics_len']['end'] = self._io.pos()
        self._debug['network_info_ofs']['start'] = self._io.pos()
        self.network_info_ofs = self._io.read_u4le()
        self._debug['network_info_ofs']['end'] = self._io.pos()
        self._debug['network_info_len']['start'] = self._io.pos()
        self.network_info_len = self._io.read_u4le()
        self._debug['network_info_len']['end'] = self._io.pos()
        self._debug['conversation_stats_ofs']['start'] = self._io.pos()
        self.conversation_stats_ofs = self._io.read_u4le()
        self._debug['conversation_stats_ofs']['end'] = self._io.pos()
        self._debug['conversation_stats_len']['start'] = self._io.pos()
        self.conversation_stats_len = self._io.read_u4le()
        self._debug['conversation_stats_len']['end'] = self._io.pos()

    class FrameIndex(KaitaiStruct):
        SEQ_FIELDS = ["entries"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['entries']['start'] = self._io.pos()
            self.entries = []
            i = 0
            while not self._io.is_eof():
                if not 'arr' in self._debug['entries']:
                    self._debug['entries']['arr'] = []
                self._debug['entries']['arr'].append({'start': self._io.pos()})
                _t_entries = MicrosoftNetworkMonitorV2.FrameIndexEntry(self._io, self, self._root)
                _t_entries._read()
                self.entries.append(_t_entries)
                self._debug['entries']['arr'][len(self.entries) - 1]['end'] = self._io.pos()
                i += 1

            self._debug['entries']['end'] = self._io.pos()


    class FrameIndexEntry(KaitaiStruct):
        """Each index entry is just a pointer to where the frame data is
        stored in the file.
        """
        SEQ_FIELDS = ["ofs"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['ofs']['start'] = self._io.pos()
            self.ofs = self._io.read_u4le()
            self._debug['ofs']['end'] = self._io.pos()

        @property
        def body(self):
            """Frame body itself."""
            if hasattr(self, '_m_body'):
                return self._m_body if hasattr(self, '_m_body') else None

            io = self._root._io
            _pos = io.pos()
            io.seek(self.ofs)
            self._debug['_m_body']['start'] = io.pos()
            self._m_body = MicrosoftNetworkMonitorV2.Frame(io, self, self._root)
            self._m_body._read()
            self._debug['_m_body']['end'] = io.pos()
            io.seek(_pos)
            return self._m_body if hasattr(self, '_m_body') else None


    class Frame(KaitaiStruct):
        """A container for actually captured network data. Allow to
        timestamp individual frames and designates how much data from
        the original packet was actually written into the file.
        
        .. seealso::
           Source - https://docs.microsoft.com/en-us/windows/win32/netmon2/frame
        """
        SEQ_FIELDS = ["ts_delta", "orig_len", "inc_len", "body"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['ts_delta']['start'] = self._io.pos()
            self.ts_delta = self._io.read_u8le()
            self._debug['ts_delta']['end'] = self._io.pos()
            self._debug['orig_len']['start'] = self._io.pos()
            self.orig_len = self._io.read_u4le()
            self._debug['orig_len']['end'] = self._io.pos()
            self._debug['inc_len']['start'] = self._io.pos()
            self.inc_len = self._io.read_u4le()
            self._debug['inc_len']['end'] = self._io.pos()
            self._debug['body']['start'] = self._io.pos()
            _on = self._root.mac_type
            if _on == MicrosoftNetworkMonitorV2.Linktype.ethernet:
                self._raw_body = self._io.read_bytes(self.inc_len)
                _io__raw_body = KaitaiStream(BytesIO(self._raw_body))
                self.body = ethernet_frame.EthernetFrame(_io__raw_body)
                self.body._read()
            else:
                self.body = self._io.read_bytes(self.inc_len)
            self._debug['body']['end'] = self._io.pos()


    @property
    def frame_table(self):
        """Index that is used to access individual captured frames."""
        if hasattr(self, '_m_frame_table'):
            return self._m_frame_table if hasattr(self, '_m_frame_table') else None

        _pos = self._io.pos()
        self._io.seek(self.frame_table_ofs)
        self._debug['_m_frame_table']['start'] = self._io.pos()
        self._raw__m_frame_table = self._io.read_bytes(self.frame_table_len)
        _io__raw__m_frame_table = KaitaiStream(BytesIO(self._raw__m_frame_table))
        self._m_frame_table = MicrosoftNetworkMonitorV2.FrameIndex(_io__raw__m_frame_table, self, self._root)
        self._m_frame_table._read()
        self._debug['_m_frame_table']['end'] = self._io.pos()
        self._io.seek(_pos)
        return self._m_frame_table if hasattr(self, '_m_frame_table') else None





# This is a generated file! Please edit source .ksy file and use kaitai-struct-compiler to rebuild

from pkg_resources import parse_version
import kaitaistruct
from kaitaistruct import KaitaiStruct, KaitaiStream, BytesIO
import collections


if parse_version(kaitaistruct.__version__) < parse_version('0.9'):
    raise Exception("Incompatible Kaitai Struct Python API: 0.9 or later is required, but you have %s" % (kaitaistruct.__version__))

class GranTurismoVol(KaitaiStruct):
    SEQ_FIELDS = ["magic", "num_files", "num_entries", "reserved", "offsets"]
    def __init__(self, _io, _parent=None, _root=None):
        self._io = _io
        self._parent = _parent
        self._root = _root if _root else self
        self._debug = collections.defaultdict(dict)

    def _read(self):
        self._debug['magic']['start'] = self._io.pos()
        self.magic = self._io.read_bytes(8)
        self._debug['magic']['end'] = self._io.pos()
        if not self.magic == b"\x47\x54\x46\x53\x00\x00\x00\x00":
            raise kaitaistruct.ValidationNotEqualError(b"\x47\x54\x46\x53\x00\x00\x00\x00", self.magic, self._io, u"/seq/0")
        self._debug['num_files']['start'] = self._io.pos()
        self.num_files = self._io.read_u2le()
        self._debug['num_files']['end'] = self._io.pos()
        self._debug['num_entries']['start'] = self._io.pos()
        self.num_entries = self._io.read_u2le()
        self._debug['num_entries']['end'] = self._io.pos()
        self._debug['reserved']['start'] = self._io.pos()
        self.reserved = self._io.read_bytes(4)
        self._debug['reserved']['end'] = self._io.pos()
        if not self.reserved == b"\x00\x00\x00\x00":
            raise kaitaistruct.ValidationNotEqualError(b"\x00\x00\x00\x00", self.reserved, self._io, u"/seq/3")
        self._debug['offsets']['start'] = self._io.pos()
        self.offsets = [None] * (self.num_files)
        for i in range(self.num_files):
            if not 'arr' in self._debug['offsets']:
                self._debug['offsets']['arr'] = []
            self._debug['offsets']['arr'].append({'start': self._io.pos()})
            self.offsets[i] = self._io.read_u4le()
            self._debug['offsets']['arr'][i]['end'] = self._io.pos()

        self._debug['offsets']['end'] = self._io.pos()

    class FileInfo(KaitaiStruct):
        SEQ_FIELDS = ["timestamp", "offset_idx", "flags", "name"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['timestamp']['start'] = self._io.pos()
            self.timestamp = self._io.read_u4le()
            self._debug['timestamp']['end'] = self._io.pos()
            self._debug['offset_idx']['start'] = self._io.pos()
            self.offset_idx = self._io.read_u2le()
            self._debug['offset_idx']['end'] = self._io.pos()
            self._debug['flags']['start'] = self._io.pos()
            self.flags = self._io.read_u1()
            self._debug['flags']['end'] = self._io.pos()
            self._debug['name']['start'] = self._io.pos()
            self.name = (KaitaiStream.bytes_terminate(KaitaiStream.bytes_strip_right(self._io.read_bytes(25), 0), 0, False)).decode(u"ASCII")
            self._debug['name']['end'] = self._io.pos()

        @property
        def size(self):
            if hasattr(self, '_m_size'):
                return self._m_size if hasattr(self, '_m_size') else None

            self._m_size = ((self._root.offsets[(self.offset_idx + 1)] & 4294965248) - self._root.offsets[self.offset_idx])
            return self._m_size if hasattr(self, '_m_size') else None

        @property
        def body(self):
            if hasattr(self, '_m_body'):
                return self._m_body if hasattr(self, '_m_body') else None

            if not (self.is_dir):
                _pos = self._io.pos()
                self._io.seek((self._root.offsets[self.offset_idx] & 4294965248))
                self._debug['_m_body']['start'] = self._io.pos()
                self._m_body = self._io.read_bytes(self.size)
                self._debug['_m_body']['end'] = self._io.pos()
                self._io.seek(_pos)

            return self._m_body if hasattr(self, '_m_body') else None

        @property
        def is_dir(self):
            if hasattr(self, '_m_is_dir'):
                return self._m_is_dir if hasattr(self, '_m_is_dir') else None

            self._m_is_dir = (self.flags & 1) != 0
            return self._m_is_dir if hasattr(self, '_m_is_dir') else None

        @property
        def is_last_entry(self):
            if hasattr(self, '_m_is_last_entry'):
                return self._m_is_last_entry if hasattr(self, '_m_is_last_entry') else None

            self._m_is_last_entry = (self.flags & 128) != 0
            return self._m_is_last_entry if hasattr(self, '_m_is_last_entry') else None


    @property
    def ofs_dir(self):
        if hasattr(self, '_m_ofs_dir'):
            return self._m_ofs_dir if hasattr(self, '_m_ofs_dir') else None

        self._m_ofs_dir = self.offsets[1]
        return self._m_ofs_dir if hasattr(self, '_m_ofs_dir') else None

    @property
    def files(self):
        if hasattr(self, '_m_files'):
            return self._m_files if hasattr(self, '_m_files') else None

        _pos = self._io.pos()
        self._io.seek((self.ofs_dir & 4294965248))
        self._debug['_m_files']['start'] = self._io.pos()
        self._m_files = [None] * (self._root.num_entries)
        for i in range(self._root.num_entries):
            if not 'arr' in self._debug['_m_files']:
                self._debug['_m_files']['arr'] = []
            self._debug['_m_files']['arr'].append({'start': self._io.pos()})
            _t__m_files = GranTurismoVol.FileInfo(self._io, self, self._root)
            _t__m_files._read()
            self._m_files[i] = _t__m_files
            self._debug['_m_files']['arr'][i]['end'] = self._io.pos()

        self._debug['_m_files']['end'] = self._io.pos()
        self._io.seek(_pos)
        return self._m_files if hasattr(self, '_m_files') else None





# This is a generated file! Please edit source .ksy file and use kaitai-struct-compiler to rebuild

from pkg_resources import parse_version
import kaitaistruct
from kaitaistruct import KaitaiStruct, KaitaiStream, BytesIO
from enum import Enum
import collections


if parse_version(kaitaistruct.__version__) < parse_version('0.9'):
    raise Exception("Incompatible Kaitai Struct Python API: 0.9 or later is required, but you have %s" % (kaitaistruct.__version__))

class MinecraftNbt(KaitaiStruct):
    """A structured binary format native to Minecraft for saving game data and transferring
    it over the network (in multiplayer), such as player data
    ([`<player>.dat`](https://minecraft.gamepedia.com/Player.dat_format); contains
    e.g. player's inventory and location), saved worlds
    ([`level.dat`](
      https://minecraft.gamepedia.com/Java_Edition_level_format#level.dat_format
    ) and [Chunk format](https://minecraft.gamepedia.com/Chunk_format#NBT_structure)),
    list of saved multiplayer servers
    ([`servers.dat`](https://minecraft.gamepedia.com/Servers.dat_format)) and so on -
    see <https://minecraft.gamepedia.com/NBT_format#Uses>.
    
    The entire file should be _gzip_-compressed (in accordance with the original
    specification [NBT.txt](
      https://web.archive.org/web/20110723210920/https://www.minecraft.net/docs/NBT.txt
    ) by Notch), but can also be compressed with _zlib_ or uncompressed.
    
    This spec can only handle uncompressed NBT data, so be sure to first detect
    what type of data you are dealing with. You can use the Unix `file` command
    to do this (`file-5.20` or later is required; older versions do not recognize
    _zlib_-compressed data and return `application/octet-stream` instead):
    
    ```shell
    file --brief --mime-type input-unknown.nbt
    ```
    
    If it says:
    
      * `application/x-gzip` or `application/gzip` (since `file-5.37`), you can decompress it by
        * `gunzip -c input-gzip.nbt > output.nbt` or
        * `python3 -c "import sys, gzip; sys.stdout.buffer.write(
          gzip.decompress(sys.stdin.buffer.read()) )" < input-gzip.nbt > output.nbt`
      * `application/zlib`, you can use
        * `openssl zlib -d -in input-zlib.nbt -out output.nbt` (does not work on most systems)
        * `python3 -c "import sys, zlib; sys.stdout.buffer.write(
          zlib.decompress(sys.stdin.buffer.read()) )" < input-zlib.nbt > output.nbt`
      * something else (especially `image/x-pcx` and `application/octet-stream`),
        it is most likely already uncompressed.
    
    The file `output.nbt` generated by one of the above commands can already be
    processed with this Kaitai Struct specification.
    
    This spec **only** implements the Java edition format. There is also
    a [Bedrock edition](https://wiki.vg/NBT#Bedrock_edition) NBT format,
    which uses little-endian encoding and has a few other differences, but it isn't
    as popular as the Java edition format.
    
    **Implementation note:** strings in `TAG_String` are incorrectly decoded with
    standard UTF-8, while they are encoded in [**Modified UTF-8**](
      https://docs.oracle.com/javase/8/docs/api/java/io/DataInput.html#modified-utf-8
    ) (MUTF-8). That's because MUTF-8 is not supported natively by most target
    languages, and thus one must use external libraries to achieve a fully-compliant
    decoder. But decoding in standard UTF-8 is still better than nothing, and
    it usually works fine.
    
    All Unicode code points with incompatible representations in MUTF-8 and UTF-8 are
    U+0000 (_NUL_), U+D800-U+DFFF (_High_ and _Low Surrogates_) and U+10000-U+10FFFF
    (all _Supplementary_ Planes; includes e.g. emoticons, pictograms).
    A _MUTF-8_-encoded string containing these code points cannot be successfully
    decoded as UTF-8. The behavior in this case depends on the target language -
    usually an exception is thrown, or the bytes that are not valid UTF-8
    are replaced or ignored.
    
    **Sample files:**
    
      * <https://wiki.vg/NBT#Download>
      * <https://github.com/twoolie/NBT/blob/f9e892e/tests/world_test/data/scoreboard.dat>
      * <https://github.com/chmod222/cNBT/tree/3f74b69/testdata>
      * <https://github.com/PistonDevelopers/hematite_nbt/tree/0b85f89/tests>
    
    .. seealso::
       Source - https://wiki.vg/NBT
    
    
    .. seealso::
       Source - https://web.archive.org/web/20110723210920/https://www.minecraft.net/docs/NBT.txt
    
    
    .. seealso::
       Source - https://minecraft.gamepedia.com/NBT_format
    """

    class Tag(Enum):
        end = 0
        byte = 1
        short = 2
        int = 3
        long = 4
        float = 5
        double = 6
        byte_array = 7
        string = 8
        list = 9
        compound = 10
        int_array = 11
        long_array = 12
    SEQ_FIELDS = ["root_check", "root"]
    def __init__(self, _io, _parent=None, _root=None):
        self._io = _io
        self._parent = _parent
        self._root = _root if _root else self
        self._debug = collections.defaultdict(dict)

    def _read(self):
        if  ((self.root_type == MinecraftNbt.Tag.end) and (False)) :
            self._debug['root_check']['start'] = self._io.pos()
            self.root_check = self._io.read_bytes(0)
            self._debug['root_check']['end'] = self._io.pos()

        self._debug['root']['start'] = self._io.pos()
        self.root = MinecraftNbt.NamedTag(self._io, self, self._root)
        self.root._read()
        self._debug['root']['end'] = self._io.pos()

    class TagLongArray(KaitaiStruct):
        SEQ_FIELDS = ["num_tags", "tags"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['num_tags']['start'] = self._io.pos()
            self.num_tags = self._io.read_s4be()
            self._debug['num_tags']['end'] = self._io.pos()
            self._debug['tags']['start'] = self._io.pos()
            self.tags = [None] * (self.num_tags)
            for i in range(self.num_tags):
                if not 'arr' in self._debug['tags']:
                    self._debug['tags']['arr'] = []
                self._debug['tags']['arr'].append({'start': self._io.pos()})
                self.tags[i] = self._io.read_s8be()
                self._debug['tags']['arr'][i]['end'] = self._io.pos()

            self._debug['tags']['end'] = self._io.pos()

        @property
        def tags_type(self):
            if hasattr(self, '_m_tags_type'):
                return self._m_tags_type if hasattr(self, '_m_tags_type') else None

            self._m_tags_type = MinecraftNbt.Tag.long
            return self._m_tags_type if hasattr(self, '_m_tags_type') else None


    class TagByteArray(KaitaiStruct):
        SEQ_FIELDS = ["len_data", "data"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['len_data']['start'] = self._io.pos()
            self.len_data = self._io.read_s4be()
            self._debug['len_data']['end'] = self._io.pos()
            self._debug['data']['start'] = self._io.pos()
            self.data = self._io.read_bytes(self.len_data)
            self._debug['data']['end'] = self._io.pos()


    class TagIntArray(KaitaiStruct):
        SEQ_FIELDS = ["num_tags", "tags"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['num_tags']['start'] = self._io.pos()
            self.num_tags = self._io.read_s4be()
            self._debug['num_tags']['end'] = self._io.pos()
            self._debug['tags']['start'] = self._io.pos()
            self.tags = [None] * (self.num_tags)
            for i in range(self.num_tags):
                if not 'arr' in self._debug['tags']:
                    self._debug['tags']['arr'] = []
                self._debug['tags']['arr'].append({'start': self._io.pos()})
                self.tags[i] = self._io.read_s4be()
                self._debug['tags']['arr'][i]['end'] = self._io.pos()

            self._debug['tags']['end'] = self._io.pos()

        @property
        def tags_type(self):
            if hasattr(self, '_m_tags_type'):
                return self._m_tags_type if hasattr(self, '_m_tags_type') else None

            self._m_tags_type = MinecraftNbt.Tag.int
            return self._m_tags_type if hasattr(self, '_m_tags_type') else None


    class TagList(KaitaiStruct):
        SEQ_FIELDS = ["tags_type", "num_tags", "tags"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['tags_type']['start'] = self._io.pos()
            self.tags_type = KaitaiStream.resolve_enum(MinecraftNbt.Tag, self._io.read_u1())
            self._debug['tags_type']['end'] = self._io.pos()
            self._debug['num_tags']['start'] = self._io.pos()
            self.num_tags = self._io.read_s4be()
            self._debug['num_tags']['end'] = self._io.pos()
            self._debug['tags']['start'] = self._io.pos()
            self.tags = [None] * (self.num_tags)
            for i in range(self.num_tags):
                if not 'arr' in self._debug['tags']:
                    self._debug['tags']['arr'] = []
                self._debug['tags']['arr'].append({'start': self._io.pos()})
                _on = self.tags_type
                if _on == MinecraftNbt.Tag.long_array:
                    if not 'arr' in self._debug['tags']:
                        self._debug['tags']['arr'] = []
                    self._debug['tags']['arr'].append({'start': self._io.pos()})
                    _t_tags = MinecraftNbt.TagLongArray(self._io, self, self._root)
                    _t_tags._read()
                    self.tags[i] = _t_tags
                    self._debug['tags']['arr'][i]['end'] = self._io.pos()
                elif _on == MinecraftNbt.Tag.compound:
                    if not 'arr' in self._debug['tags']:
                        self._debug['tags']['arr'] = []
                    self._debug['tags']['arr'].append({'start': self._io.pos()})
                    _t_tags = MinecraftNbt.TagCompound(self._io, self, self._root)
                    _t_tags._read()
                    self.tags[i] = _t_tags
                    self._debug['tags']['arr'][i]['end'] = self._io.pos()
                elif _on == MinecraftNbt.Tag.double:
                    if not 'arr' in self._debug['tags']:
                        self._debug['tags']['arr'] = []
                    self._debug['tags']['arr'].append({'start': self._io.pos()})
                    self.tags[i] = self._io.read_f8be()
                    self._debug['tags']['arr'][i]['end'] = self._io.pos()
                elif _on == MinecraftNbt.Tag.list:
                    if not 'arr' in self._debug['tags']:
                        self._debug['tags']['arr'] = []
                    self._debug['tags']['arr'].append({'start': self._io.pos()})
                    _t_tags = MinecraftNbt.TagList(self._io, self, self._root)
                    _t_tags._read()
                    self.tags[i] = _t_tags
                    self._debug['tags']['arr'][i]['end'] = self._io.pos()
                elif _on == MinecraftNbt.Tag.float:
                    if not 'arr' in self._debug['tags']:
                        self._debug['tags']['arr'] = []
                    self._debug['tags']['arr'].append({'start': self._io.pos()})
                    self.tags[i] = self._io.read_f4be()
                    self._debug['tags']['arr'][i]['end'] = self._io.pos()
                elif _on == MinecraftNbt.Tag.short:
                    if not 'arr' in self._debug['tags']:
                        self._debug['tags']['arr'] = []
                    self._debug['tags']['arr'].append({'start': self._io.pos()})
                    self.tags[i] = self._io.read_s2be()
                    self._debug['tags']['arr'][i]['end'] = self._io.pos()
                elif _on == MinecraftNbt.Tag.int:
                    if not 'arr' in self._debug['tags']:
                        self._debug['tags']['arr'] = []
                    self._debug['tags']['arr'].append({'start': self._io.pos()})
                    self.tags[i] = self._io.read_s4be()
                    self._debug['tags']['arr'][i]['end'] = self._io.pos()
                elif _on == MinecraftNbt.Tag.byte_array:
                    if not 'arr' in self._debug['tags']:
                        self._debug['tags']['arr'] = []
                    self._debug['tags']['arr'].append({'start': self._io.pos()})
                    _t_tags = MinecraftNbt.TagByteArray(self._io, self, self._root)
                    _t_tags._read()
                    self.tags[i] = _t_tags
                    self._debug['tags']['arr'][i]['end'] = self._io.pos()
                elif _on == MinecraftNbt.Tag.byte:
                    if not 'arr' in self._debug['tags']:
                        self._debug['tags']['arr'] = []
                    self._debug['tags']['arr'].append({'start': self._io.pos()})
                    self.tags[i] = self._io.read_s1()
                    self._debug['tags']['arr'][i]['end'] = self._io.pos()
                elif _on == MinecraftNbt.Tag.int_array:
                    if not 'arr' in self._debug['tags']:
                        self._debug['tags']['arr'] = []
                    self._debug['tags']['arr'].append({'start': self._io.pos()})
                    _t_tags = MinecraftNbt.TagIntArray(self._io, self, self._root)
                    _t_tags._read()
                    self.tags[i] = _t_tags
                    self._debug['tags']['arr'][i]['end'] = self._io.pos()
                elif _on == MinecraftNbt.Tag.string:
                    if not 'arr' in self._debug['tags']:
                        self._debug['tags']['arr'] = []
                    self._debug['tags']['arr'].append({'start': self._io.pos()})
                    _t_tags = MinecraftNbt.TagString(self._io, self, self._root)
                    _t_tags._read()
                    self.tags[i] = _t_tags
                    self._debug['tags']['arr'][i]['end'] = self._io.pos()
                elif _on == MinecraftNbt.Tag.long:
                    if not 'arr' in self._debug['tags']:
                        self._debug['tags']['arr'] = []
                    self._debug['tags']['arr'].append({'start': self._io.pos()})
                    self.tags[i] = self._io.read_s8be()
                    self._debug['tags']['arr'][i]['end'] = self._io.pos()
                self._debug['tags']['arr'][i]['end'] = self._io.pos()

            self._debug['tags']['end'] = self._io.pos()


    class TagString(KaitaiStruct):
        SEQ_FIELDS = ["len_data", "data"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['len_data']['start'] = self._io.pos()
            self.len_data = self._io.read_u2be()
            self._debug['len_data']['end'] = self._io.pos()
            self._debug['data']['start'] = self._io.pos()
            self.data = (self._io.read_bytes(self.len_data)).decode(u"utf-8")
            self._debug['data']['end'] = self._io.pos()


    class TagCompound(KaitaiStruct):
        SEQ_FIELDS = ["tags"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['tags']['start'] = self._io.pos()
            self.tags = []
            i = 0
            while True:
                if not 'arr' in self._debug['tags']:
                    self._debug['tags']['arr'] = []
                self._debug['tags']['arr'].append({'start': self._io.pos()})
                _t_tags = MinecraftNbt.NamedTag(self._io, self, self._root)
                _t_tags._read()
                _ = _t_tags
                self.tags.append(_)
                self._debug['tags']['arr'][len(self.tags) - 1]['end'] = self._io.pos()
                if _.is_tag_end:
                    break
                i += 1
            self._debug['tags']['end'] = self._io.pos()

        @property
        def dump_num_tags(self):
            if hasattr(self, '_m_dump_num_tags'):
                return self._m_dump_num_tags if hasattr(self, '_m_dump_num_tags') else None

            self._m_dump_num_tags = (len(self.tags) - (1 if  ((len(self.tags) >= 1) and (self.tags[-1].is_tag_end))  else 0))
            return self._m_dump_num_tags if hasattr(self, '_m_dump_num_tags') else None


    class NamedTag(KaitaiStruct):
        SEQ_FIELDS = ["type", "name", "payload"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['type']['start'] = self._io.pos()
            self.type = KaitaiStream.resolve_enum(MinecraftNbt.Tag, self._io.read_u1())
            self._debug['type']['end'] = self._io.pos()
            if not (self.is_tag_end):
                self._debug['name']['start'] = self._io.pos()
                self.name = MinecraftNbt.TagString(self._io, self, self._root)
                self.name._read()
                self._debug['name']['end'] = self._io.pos()

            if not (self.is_tag_end):
                self._debug['payload']['start'] = self._io.pos()
                _on = self.type
                if _on == MinecraftNbt.Tag.long_array:
                    self.payload = MinecraftNbt.TagLongArray(self._io, self, self._root)
                    self.payload._read()
                elif _on == MinecraftNbt.Tag.compound:
                    self.payload = MinecraftNbt.TagCompound(self._io, self, self._root)
                    self.payload._read()
                elif _on == MinecraftNbt.Tag.double:
                    self.payload = self._io.read_f8be()
                elif _on == MinecraftNbt.Tag.list:
                    self.payload = MinecraftNbt.TagList(self._io, self, self._root)
                    self.payload._read()
                elif _on == MinecraftNbt.Tag.float:
                    self.payload = self._io.read_f4be()
                elif _on == MinecraftNbt.Tag.short:
                    self.payload = self._io.read_s2be()
                elif _on == MinecraftNbt.Tag.int:
                    self.payload = self._io.read_s4be()
                elif _on == MinecraftNbt.Tag.byte_array:
                    self.payload = MinecraftNbt.TagByteArray(self._io, self, self._root)
                    self.payload._read()
                elif _on == MinecraftNbt.Tag.byte:
                    self.payload = self._io.read_s1()
                elif _on == MinecraftNbt.Tag.int_array:
                    self.payload = MinecraftNbt.TagIntArray(self._io, self, self._root)
                    self.payload._read()
                elif _on == MinecraftNbt.Tag.string:
                    self.payload = MinecraftNbt.TagString(self._io, self, self._root)
                    self.payload._read()
                elif _on == MinecraftNbt.Tag.long:
                    self.payload = self._io.read_s8be()
                self._debug['payload']['end'] = self._io.pos()


        @property
        def is_tag_end(self):
            if hasattr(self, '_m_is_tag_end'):
                return self._m_is_tag_end if hasattr(self, '_m_is_tag_end') else None

            self._m_is_tag_end = self.type == MinecraftNbt.Tag.end
            return self._m_is_tag_end if hasattr(self, '_m_is_tag_end') else None


    @property
    def root_type(self):
        if hasattr(self, '_m_root_type'):
            return self._m_root_type if hasattr(self, '_m_root_type') else None

        _pos = self._io.pos()
        self._io.seek(0)
        self._debug['_m_root_type']['start'] = self._io.pos()
        self._m_root_type = KaitaiStream.resolve_enum(MinecraftNbt.Tag, self._io.read_u1())
        self._debug['_m_root_type']['end'] = self._io.pos()
        self._io.seek(_pos)
        if not self.root_type == MinecraftNbt.Tag.compound:
            raise kaitaistruct.ValidationNotEqualError(MinecraftNbt.Tag.compound, self.root_type, self._io, u"/instances/root_type")
        return self._m_root_type if hasattr(self, '_m_root_type') else None





# This is a generated file! Please edit source .ksy file and use kaitai-struct-compiler to rebuild

from pkg_resources import parse_version
import kaitaistruct
from kaitaistruct import KaitaiStruct, KaitaiStream, BytesIO
from enum import Enum
import collections


if parse_version(kaitaistruct.__version__) < parse_version('0.9'):
    raise Exception("Incompatible Kaitai Struct Python API: 0.9 or later is required, but you have %s" % (kaitaistruct.__version__))

class ZxSpectrumTap(KaitaiStruct):
    """TAP files are used by emulators of ZX Spectrum computer (released in
    1982 by Sinclair Research). TAP file stores blocks of data as if
    they are written to magnetic tape, which was used as primary media
    for ZX Spectrum. Contents of this file can be viewed as a very
    simple linear filesystem, storing named files with some basic
    metainformation prepended as a header.
    
    .. seealso::
       Source - https://faqwiki.zxnet.co.uk/wiki/TAP_format
    """

    class FlagEnum(Enum):
        header = 0
        data = 255

    class HeaderTypeEnum(Enum):
        program = 0
        num_array = 1
        char_array = 2
        bytes = 3
    SEQ_FIELDS = ["blocks"]
    def __init__(self, _io, _parent=None, _root=None):
        self._io = _io
        self._parent = _parent
        self._root = _root if _root else self
        self._debug = collections.defaultdict(dict)

    def _read(self):
        self._debug['blocks']['start'] = self._io.pos()
        self.blocks = []
        i = 0
        while not self._io.is_eof():
            if not 'arr' in self._debug['blocks']:
                self._debug['blocks']['arr'] = []
            self._debug['blocks']['arr'].append({'start': self._io.pos()})
            _t_blocks = ZxSpectrumTap.Block(self._io, self, self._root)
            _t_blocks._read()
            self.blocks.append(_t_blocks)
            self._debug['blocks']['arr'][len(self.blocks) - 1]['end'] = self._io.pos()
            i += 1

        self._debug['blocks']['end'] = self._io.pos()

    class Block(KaitaiStruct):
        SEQ_FIELDS = ["len_block", "flag", "header", "data", "headerless_data"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['len_block']['start'] = self._io.pos()
            self.len_block = self._io.read_u2le()
            self._debug['len_block']['end'] = self._io.pos()
            self._debug['flag']['start'] = self._io.pos()
            self.flag = KaitaiStream.resolve_enum(ZxSpectrumTap.FlagEnum, self._io.read_u1())
            self._debug['flag']['end'] = self._io.pos()
            if  ((self.len_block == 19) and (self.flag == ZxSpectrumTap.FlagEnum.header)) :
                self._debug['header']['start'] = self._io.pos()
                self.header = ZxSpectrumTap.Header(self._io, self, self._root)
                self.header._read()
                self._debug['header']['end'] = self._io.pos()

            if self.len_block == 19:
                self._debug['data']['start'] = self._io.pos()
                self.data = self._io.read_bytes((self.header.len_data + 4))
                self._debug['data']['end'] = self._io.pos()

            if self.flag == ZxSpectrumTap.FlagEnum.data:
                self._debug['headerless_data']['start'] = self._io.pos()
                self.headerless_data = self._io.read_bytes((self.len_block - 1))
                self._debug['headerless_data']['end'] = self._io.pos()



    class ProgramParams(KaitaiStruct):
        SEQ_FIELDS = ["autostart_line", "len_program"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['autostart_line']['start'] = self._io.pos()
            self.autostart_line = self._io.read_u2le()
            self._debug['autostart_line']['end'] = self._io.pos()
            self._debug['len_program']['start'] = self._io.pos()
            self.len_program = self._io.read_u2le()
            self._debug['len_program']['end'] = self._io.pos()


    class BytesParams(KaitaiStruct):
        SEQ_FIELDS = ["start_address", "reserved"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['start_address']['start'] = self._io.pos()
            self.start_address = self._io.read_u2le()
            self._debug['start_address']['end'] = self._io.pos()
            self._debug['reserved']['start'] = self._io.pos()
            self.reserved = self._io.read_bytes(2)
            self._debug['reserved']['end'] = self._io.pos()


    class Header(KaitaiStruct):
        SEQ_FIELDS = ["header_type", "filename", "len_data", "params", "checksum"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['header_type']['start'] = self._io.pos()
            self.header_type = KaitaiStream.resolve_enum(ZxSpectrumTap.HeaderTypeEnum, self._io.read_u1())
            self._debug['header_type']['end'] = self._io.pos()
            self._debug['filename']['start'] = self._io.pos()
            self.filename = KaitaiStream.bytes_strip_right(self._io.read_bytes(10), 32)
            self._debug['filename']['end'] = self._io.pos()
            self._debug['len_data']['start'] = self._io.pos()
            self.len_data = self._io.read_u2le()
            self._debug['len_data']['end'] = self._io.pos()
            self._debug['params']['start'] = self._io.pos()
            _on = self.header_type
            if _on == ZxSpectrumTap.HeaderTypeEnum.program:
                self.params = ZxSpectrumTap.ProgramParams(self._io, self, self._root)
                self.params._read()
            elif _on == ZxSpectrumTap.HeaderTypeEnum.num_array:
                self.params = ZxSpectrumTap.ArrayParams(self._io, self, self._root)
                self.params._read()
            elif _on == ZxSpectrumTap.HeaderTypeEnum.char_array:
                self.params = ZxSpectrumTap.ArrayParams(self._io, self, self._root)
                self.params._read()
            elif _on == ZxSpectrumTap.HeaderTypeEnum.bytes:
                self.params = ZxSpectrumTap.BytesParams(self._io, self, self._root)
                self.params._read()
            self._debug['params']['end'] = self._io.pos()
            self._debug['checksum']['start'] = self._io.pos()
            self.checksum = self._io.read_u1()
            self._debug['checksum']['end'] = self._io.pos()


    class ArrayParams(KaitaiStruct):
        SEQ_FIELDS = ["reserved", "var_name", "reserved1"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['reserved']['start'] = self._io.pos()
            self.reserved = self._io.read_u1()
            self._debug['reserved']['end'] = self._io.pos()
            self._debug['var_name']['start'] = self._io.pos()
            self.var_name = self._io.read_u1()
            self._debug['var_name']['end'] = self._io.pos()
            self._debug['reserved1']['start'] = self._io.pos()
            self.reserved1 = self._io.read_bytes(2)
            self._debug['reserved1']['end'] = self._io.pos()
            if not self.reserved1 == b"\x00\x80":
                raise kaitaistruct.ValidationNotEqualError(b"\x00\x80", self.reserved1, self._io, u"/types/array_params/seq/2")






# This is a generated file! Please edit source .ksy file and use kaitai-struct-compiler to rebuild

from pkg_resources import parse_version
import kaitaistruct
from kaitaistruct import KaitaiStruct, KaitaiStream, BytesIO
import collections


if parse_version(kaitaistruct.__version__) < parse_version('0.9'):
    raise Exception("Incompatible Kaitai Struct Python API: 0.9 or later is required, but you have %s" % (kaitaistruct.__version__))

class AndroidImg(KaitaiStruct):
    """
    .. seealso::
       Source - https://source.android.com/devices/bootloader/boot-image-header
    """
    SEQ_FIELDS = ["magic", "kernel", "ramdisk", "second", "tags_load", "page_size", "header_version", "os_version", "name", "cmdline", "sha", "extra_cmdline", "recovery_dtbo", "boot_header_size", "dtb"]
    def __init__(self, _io, _parent=None, _root=None):
        self._io = _io
        self._parent = _parent
        self._root = _root if _root else self
        self._debug = collections.defaultdict(dict)

    def _read(self):
        self._debug['magic']['start'] = self._io.pos()
        self.magic = self._io.read_bytes(8)
        self._debug['magic']['end'] = self._io.pos()
        if not self.magic == b"\x41\x4E\x44\x52\x4F\x49\x44\x21":
            raise kaitaistruct.ValidationNotEqualError(b"\x41\x4E\x44\x52\x4F\x49\x44\x21", self.magic, self._io, u"/seq/0")
        self._debug['kernel']['start'] = self._io.pos()
        self.kernel = AndroidImg.Load(self._io, self, self._root)
        self.kernel._read()
        self._debug['kernel']['end'] = self._io.pos()
        self._debug['ramdisk']['start'] = self._io.pos()
        self.ramdisk = AndroidImg.Load(self._io, self, self._root)
        self.ramdisk._read()
        self._debug['ramdisk']['end'] = self._io.pos()
        self._debug['second']['start'] = self._io.pos()
        self.second = AndroidImg.Load(self._io, self, self._root)
        self.second._read()
        self._debug['second']['end'] = self._io.pos()
        self._debug['tags_load']['start'] = self._io.pos()
        self.tags_load = self._io.read_u4le()
        self._debug['tags_load']['end'] = self._io.pos()
        self._debug['page_size']['start'] = self._io.pos()
        self.page_size = self._io.read_u4le()
        self._debug['page_size']['end'] = self._io.pos()
        self._debug['header_version']['start'] = self._io.pos()
        self.header_version = self._io.read_u4le()
        self._debug['header_version']['end'] = self._io.pos()
        self._debug['os_version']['start'] = self._io.pos()
        self.os_version = AndroidImg.OsVersion(self._io, self, self._root)
        self.os_version._read()
        self._debug['os_version']['end'] = self._io.pos()
        self._debug['name']['start'] = self._io.pos()
        self.name = (KaitaiStream.bytes_terminate(self._io.read_bytes(16), 0, False)).decode(u"ASCII")
        self._debug['name']['end'] = self._io.pos()
        self._debug['cmdline']['start'] = self._io.pos()
        self.cmdline = (KaitaiStream.bytes_terminate(self._io.read_bytes(512), 0, False)).decode(u"ASCII")
        self._debug['cmdline']['end'] = self._io.pos()
        self._debug['sha']['start'] = self._io.pos()
        self.sha = self._io.read_bytes(32)
        self._debug['sha']['end'] = self._io.pos()
        self._debug['extra_cmdline']['start'] = self._io.pos()
        self.extra_cmdline = (KaitaiStream.bytes_terminate(self._io.read_bytes(1024), 0, False)).decode(u"ASCII")
        self._debug['extra_cmdline']['end'] = self._io.pos()
        if self.header_version > 0:
            self._debug['recovery_dtbo']['start'] = self._io.pos()
            self.recovery_dtbo = AndroidImg.SizeOffset(self._io, self, self._root)
            self.recovery_dtbo._read()
            self._debug['recovery_dtbo']['end'] = self._io.pos()

        if self.header_version > 0:
            self._debug['boot_header_size']['start'] = self._io.pos()
            self.boot_header_size = self._io.read_u4le()
            self._debug['boot_header_size']['end'] = self._io.pos()

        if self.header_version > 1:
            self._debug['dtb']['start'] = self._io.pos()
            self.dtb = AndroidImg.LoadLong(self._io, self, self._root)
            self.dtb._read()
            self._debug['dtb']['end'] = self._io.pos()


    class Load(KaitaiStruct):
        SEQ_FIELDS = ["size", "addr"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['size']['start'] = self._io.pos()
            self.size = self._io.read_u4le()
            self._debug['size']['end'] = self._io.pos()
            self._debug['addr']['start'] = self._io.pos()
            self.addr = self._io.read_u4le()
            self._debug['addr']['end'] = self._io.pos()


    class LoadLong(KaitaiStruct):
        SEQ_FIELDS = ["size", "addr"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['size']['start'] = self._io.pos()
            self.size = self._io.read_u4le()
            self._debug['size']['end'] = self._io.pos()
            self._debug['addr']['start'] = self._io.pos()
            self.addr = self._io.read_u8le()
            self._debug['addr']['end'] = self._io.pos()


    class SizeOffset(KaitaiStruct):
        SEQ_FIELDS = ["size", "offset"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['size']['start'] = self._io.pos()
            self.size = self._io.read_u4le()
            self._debug['size']['end'] = self._io.pos()
            self._debug['offset']['start'] = self._io.pos()
            self.offset = self._io.read_u8le()
            self._debug['offset']['end'] = self._io.pos()


    class OsVersion(KaitaiStruct):
        SEQ_FIELDS = ["version"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['version']['start'] = self._io.pos()
            self.version = self._io.read_u4le()
            self._debug['version']['end'] = self._io.pos()

        @property
        def month(self):
            if hasattr(self, '_m_month'):
                return self._m_month if hasattr(self, '_m_month') else None

            self._m_month = (self.version & 15)
            return self._m_month if hasattr(self, '_m_month') else None

        @property
        def patch(self):
            if hasattr(self, '_m_patch'):
                return self._m_patch if hasattr(self, '_m_patch') else None

            self._m_patch = ((self.version >> 11) & 127)
            return self._m_patch if hasattr(self, '_m_patch') else None

        @property
        def year(self):
            if hasattr(self, '_m_year'):
                return self._m_year if hasattr(self, '_m_year') else None

            self._m_year = (((self.version >> 4) & 127) + 2000)
            return self._m_year if hasattr(self, '_m_year') else None

        @property
        def major(self):
            if hasattr(self, '_m_major'):
                return self._m_major if hasattr(self, '_m_major') else None

            self._m_major = ((self.version >> 25) & 127)
            return self._m_major if hasattr(self, '_m_major') else None

        @property
        def minor(self):
            if hasattr(self, '_m_minor'):
                return self._m_minor if hasattr(self, '_m_minor') else None

            self._m_minor = ((self.version >> 18) & 127)
            return self._m_minor if hasattr(self, '_m_minor') else None


    @property
    def kernel_img(self):
        if hasattr(self, '_m_kernel_img'):
            return self._m_kernel_img if hasattr(self, '_m_kernel_img') else None

        _pos = self._io.pos()
        self._io.seek(self.page_size)
        self._debug['_m_kernel_img']['start'] = self._io.pos()
        self._m_kernel_img = self._io.read_bytes(self.kernel.size)
        self._debug['_m_kernel_img']['end'] = self._io.pos()
        self._io.seek(_pos)
        return self._m_kernel_img if hasattr(self, '_m_kernel_img') else None

    @property
    def tags_offset(self):
        """tags offset from base."""
        if hasattr(self, '_m_tags_offset'):
            return self._m_tags_offset if hasattr(self, '_m_tags_offset') else None

        self._m_tags_offset = (self.tags_load - self.base)
        return self._m_tags_offset if hasattr(self, '_m_tags_offset') else None

    @property
    def ramdisk_offset(self):
        """ramdisk offset from base."""
        if hasattr(self, '_m_ramdisk_offset'):
            return self._m_ramdisk_offset if hasattr(self, '_m_ramdisk_offset') else None

        self._m_ramdisk_offset = ((self.ramdisk.addr - self.base) if self.ramdisk.addr > 0 else 0)
        return self._m_ramdisk_offset if hasattr(self, '_m_ramdisk_offset') else None

    @property
    def second_offset(self):
        """2nd bootloader offset from base."""
        if hasattr(self, '_m_second_offset'):
            return self._m_second_offset if hasattr(self, '_m_second_offset') else None

        self._m_second_offset = ((self.second.addr - self.base) if self.second.addr > 0 else 0)
        return self._m_second_offset if hasattr(self, '_m_second_offset') else None

    @property
    def kernel_offset(self):
        """kernel offset from base."""
        if hasattr(self, '_m_kernel_offset'):
            return self._m_kernel_offset if hasattr(self, '_m_kernel_offset') else None

        self._m_kernel_offset = (self.kernel.addr - self.base)
        return self._m_kernel_offset if hasattr(self, '_m_kernel_offset') else None

    @property
    def dtb_offset(self):
        """dtb offset from base."""
        if hasattr(self, '_m_dtb_offset'):
            return self._m_dtb_offset if hasattr(self, '_m_dtb_offset') else None

        if self.header_version > 1:
            self._m_dtb_offset = ((self.dtb.addr - self.base) if self.dtb.addr > 0 else 0)

        return self._m_dtb_offset if hasattr(self, '_m_dtb_offset') else None

    @property
    def dtb_img(self):
        if hasattr(self, '_m_dtb_img'):
            return self._m_dtb_img if hasattr(self, '_m_dtb_img') else None

        if  ((self.header_version > 1) and (self.dtb.size > 0)) :
            _pos = self._io.pos()
            self._io.seek((((((((self.page_size + self.kernel.size) + self.ramdisk.size) + self.second.size) + self.recovery_dtbo.size) + self.page_size) - 1) // self.page_size * self.page_size))
            self._debug['_m_dtb_img']['start'] = self._io.pos()
            self._m_dtb_img = self._io.read_bytes(self.dtb.size)
            self._debug['_m_dtb_img']['end'] = self._io.pos()
            self._io.seek(_pos)

        return self._m_dtb_img if hasattr(self, '_m_dtb_img') else None

    @property
    def ramdisk_img(self):
        if hasattr(self, '_m_ramdisk_img'):
            return self._m_ramdisk_img if hasattr(self, '_m_ramdisk_img') else None

        if self.ramdisk.size > 0:
            _pos = self._io.pos()
            self._io.seek(((((self.page_size + self.kernel.size) + self.page_size) - 1) // self.page_size * self.page_size))
            self._debug['_m_ramdisk_img']['start'] = self._io.pos()
            self._m_ramdisk_img = self._io.read_bytes(self.ramdisk.size)
            self._debug['_m_ramdisk_img']['end'] = self._io.pos()
            self._io.seek(_pos)

        return self._m_ramdisk_img if hasattr(self, '_m_ramdisk_img') else None

    @property
    def recovery_dtbo_img(self):
        if hasattr(self, '_m_recovery_dtbo_img'):
            return self._m_recovery_dtbo_img if hasattr(self, '_m_recovery_dtbo_img') else None

        if  ((self.header_version > 0) and (self.recovery_dtbo.size > 0)) :
            _pos = self._io.pos()
            self._io.seek(self.recovery_dtbo.offset)
            self._debug['_m_recovery_dtbo_img']['start'] = self._io.pos()
            self._m_recovery_dtbo_img = self._io.read_bytes(self.recovery_dtbo.size)
            self._debug['_m_recovery_dtbo_img']['end'] = self._io.pos()
            self._io.seek(_pos)

        return self._m_recovery_dtbo_img if hasattr(self, '_m_recovery_dtbo_img') else None

    @property
    def second_img(self):
        if hasattr(self, '_m_second_img'):
            return self._m_second_img if hasattr(self, '_m_second_img') else None

        if self.second.size > 0:
            _pos = self._io.pos()
            self._io.seek((((((self.page_size + self.kernel.size) + self.ramdisk.size) + self.page_size) - 1) // self.page_size * self.page_size))
            self._debug['_m_second_img']['start'] = self._io.pos()
            self._m_second_img = self._io.read_bytes(self.second.size)
            self._debug['_m_second_img']['end'] = self._io.pos()
            self._io.seek(_pos)

        return self._m_second_img if hasattr(self, '_m_second_img') else None

    @property
    def base(self):
        """base loading address."""
        if hasattr(self, '_m_base'):
            return self._m_base if hasattr(self, '_m_base') else None

        self._m_base = (self.kernel.addr - 32768)
        return self._m_base if hasattr(self, '_m_base') else None





# This is a generated file! Please edit source .ksy file and use kaitai-struct-compiler to rebuild

from pkg_resources import parse_version
import kaitaistruct
from kaitaistruct import KaitaiStruct, KaitaiStream, BytesIO
from enum import Enum
import collections
import zlib


if parse_version(kaitaistruct.__version__) < parse_version('0.9'):
    raise Exception("Incompatible Kaitai Struct Python API: 0.9 or later is required, but you have %s" % (kaitaistruct.__version__))

class Fallout2Dat(KaitaiStruct):

    class Compression(Enum):
        none = 0
        zlib = 1
    SEQ_FIELDS = []
    def __init__(self, _io, _parent=None, _root=None):
        self._io = _io
        self._parent = _parent
        self._root = _root if _root else self
        self._debug = collections.defaultdict(dict)

    def _read(self):
        pass

    class Pstr(KaitaiStruct):
        SEQ_FIELDS = ["size", "str"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['size']['start'] = self._io.pos()
            self.size = self._io.read_u4le()
            self._debug['size']['end'] = self._io.pos()
            self._debug['str']['start'] = self._io.pos()
            self.str = (self._io.read_bytes(self.size)).decode(u"ASCII")
            self._debug['str']['end'] = self._io.pos()


    class Footer(KaitaiStruct):
        SEQ_FIELDS = ["index_size", "file_size"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['index_size']['start'] = self._io.pos()
            self.index_size = self._io.read_u4le()
            self._debug['index_size']['end'] = self._io.pos()
            self._debug['file_size']['start'] = self._io.pos()
            self.file_size = self._io.read_u4le()
            self._debug['file_size']['end'] = self._io.pos()


    class Index(KaitaiStruct):
        SEQ_FIELDS = ["file_count", "files"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['file_count']['start'] = self._io.pos()
            self.file_count = self._io.read_u4le()
            self._debug['file_count']['end'] = self._io.pos()
            self._debug['files']['start'] = self._io.pos()
            self.files = [None] * (self.file_count)
            for i in range(self.file_count):
                if not 'arr' in self._debug['files']:
                    self._debug['files']['arr'] = []
                self._debug['files']['arr'].append({'start': self._io.pos()})
                _t_files = Fallout2Dat.File(self._io, self, self._root)
                _t_files._read()
                self.files[i] = _t_files
                self._debug['files']['arr'][i]['end'] = self._io.pos()

            self._debug['files']['end'] = self._io.pos()


    class File(KaitaiStruct):
        SEQ_FIELDS = ["name", "flags", "size_unpacked", "size_packed", "offset"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['name']['start'] = self._io.pos()
            self.name = Fallout2Dat.Pstr(self._io, self, self._root)
            self.name._read()
            self._debug['name']['end'] = self._io.pos()
            self._debug['flags']['start'] = self._io.pos()
            self.flags = KaitaiStream.resolve_enum(Fallout2Dat.Compression, self._io.read_u1())
            self._debug['flags']['end'] = self._io.pos()
            self._debug['size_unpacked']['start'] = self._io.pos()
            self.size_unpacked = self._io.read_u4le()
            self._debug['size_unpacked']['end'] = self._io.pos()
            self._debug['size_packed']['start'] = self._io.pos()
            self.size_packed = self._io.read_u4le()
            self._debug['size_packed']['end'] = self._io.pos()
            self._debug['offset']['start'] = self._io.pos()
            self.offset = self._io.read_u4le()
            self._debug['offset']['end'] = self._io.pos()

        @property
        def contents_raw(self):
            if hasattr(self, '_m_contents_raw'):
                return self._m_contents_raw if hasattr(self, '_m_contents_raw') else None

            if self.flags == Fallout2Dat.Compression.none:
                io = self._root._io
                _pos = io.pos()
                io.seek(self.offset)
                self._debug['_m_contents_raw']['start'] = io.pos()
                self._m_contents_raw = io.read_bytes(self.size_unpacked)
                self._debug['_m_contents_raw']['end'] = io.pos()
                io.seek(_pos)

            return self._m_contents_raw if hasattr(self, '_m_contents_raw') else None

        @property
        def contents_zlib(self):
            if hasattr(self, '_m_contents_zlib'):
                return self._m_contents_zlib if hasattr(self, '_m_contents_zlib') else None

            if self.flags == Fallout2Dat.Compression.zlib:
                io = self._root._io
                _pos = io.pos()
                io.seek(self.offset)
                self._debug['_m_contents_zlib']['start'] = io.pos()
                self._raw__m_contents_zlib = io.read_bytes(self.size_packed)
                self._m_contents_zlib = zlib.decompress(self._raw__m_contents_zlib)
                self._debug['_m_contents_zlib']['end'] = io.pos()
                io.seek(_pos)

            return self._m_contents_zlib if hasattr(self, '_m_contents_zlib') else None

        @property
        def contents(self):
            if hasattr(self, '_m_contents'):
                return self._m_contents if hasattr(self, '_m_contents') else None

            if  ((self.flags == Fallout2Dat.Compression.zlib) or (self.flags == Fallout2Dat.Compression.none)) :
                self._m_contents = (self.contents_zlib if self.flags == Fallout2Dat.Compression.zlib else self.contents_raw)

            return self._m_contents if hasattr(self, '_m_contents') else None


    @property
    def footer(self):
        if hasattr(self, '_m_footer'):
            return self._m_footer if hasattr(self, '_m_footer') else None

        _pos = self._io.pos()
        self._io.seek((self._io.size() - 8))
        self._debug['_m_footer']['start'] = self._io.pos()
        self._m_footer = Fallout2Dat.Footer(self._io, self, self._root)
        self._m_footer._read()
        self._debug['_m_footer']['end'] = self._io.pos()
        self._io.seek(_pos)
        return self._m_footer if hasattr(self, '_m_footer') else None

    @property
    def index(self):
        if hasattr(self, '_m_index'):
            return self._m_index if hasattr(self, '_m_index') else None

        _pos = self._io.pos()
        self._io.seek(((self._io.size() - 8) - self.footer.index_size))
        self._debug['_m_index']['start'] = self._io.pos()
        self._m_index = Fallout2Dat.Index(self._io, self, self._root)
        self._m_index._read()
        self._debug['_m_index']['end'] = self._io.pos()
        self._io.seek(_pos)
        return self._m_index if hasattr(self, '_m_index') else None





# This is a generated file! Please edit source .ksy file and use kaitai-struct-compiler to rebuild

from pkg_resources import parse_version
import kaitaistruct
from kaitaistruct import KaitaiStruct, KaitaiStream, BytesIO
import collections
from enum import Enum


if parse_version(kaitaistruct.__version__) < parse_version('0.9'):
    raise Exception("Incompatible Kaitai Struct Python API: 0.9 or later is required, but you have %s" % (kaitaistruct.__version__))

class WindowsEvtLog(KaitaiStruct):
    """EVT files are Windows Event Log files written by older Windows
    operating systems (2000, XP, 2003). They are used as binary log
    files by several major Windows subsystems and
    applications. Typically, several of them can be found in
    `%WINDIR%\system32\config` directory:
    
    * Application = `AppEvent.evt`
    * System = `SysEvent.evt`
    * Security = `SecEvent.evt`
    
    Alternatively, one can export any system event log as distinct .evt
    file using relevant option in Event Viewer application.
    
    A Windows application can submit an entry into these logs using
    [ReportEventA](https://docs.microsoft.com/en-us/windows/win32/api/winbase/nf-winbase-reporteventa)
    function of Windows API.
    
    Internally, EVT files consist of a fixed-size header and event
    records. There are several usage scenarios (non-wrapping vs wrapping
    log files) which result in slightly different organization of
    records.
    
    .. seealso::
       Source - https://docs.microsoft.com/en-us/windows/win32/eventlog/event-log-file-format
    """
    SEQ_FIELDS = ["header", "records"]
    def __init__(self, _io, _parent=None, _root=None):
        self._io = _io
        self._parent = _parent
        self._root = _root if _root else self
        self._debug = collections.defaultdict(dict)

    def _read(self):
        self._debug['header']['start'] = self._io.pos()
        self.header = WindowsEvtLog.Header(self._io, self, self._root)
        self.header._read()
        self._debug['header']['end'] = self._io.pos()
        self._debug['records']['start'] = self._io.pos()
        self.records = []
        i = 0
        while not self._io.is_eof():
            if not 'arr' in self._debug['records']:
                self._debug['records']['arr'] = []
            self._debug['records']['arr'].append({'start': self._io.pos()})
            _t_records = WindowsEvtLog.Record(self._io, self, self._root)
            _t_records._read()
            self.records.append(_t_records)
            self._debug['records']['arr'][len(self.records) - 1]['end'] = self._io.pos()
            i += 1

        self._debug['records']['end'] = self._io.pos()

    class Header(KaitaiStruct):
        """
        .. seealso::
           Source - https://docs.microsoft.com/en-us/previous-versions/windows/desktop/legacy/bb309024(v=vs.85)
        """
        SEQ_FIELDS = ["len_header", "magic", "version_major", "version_minor", "ofs_start", "ofs_end", "cur_rec_idx", "oldest_rec_idx", "len_file_max", "flags", "retention", "len_header_2"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['len_header']['start'] = self._io.pos()
            self.len_header = self._io.read_u4le()
            self._debug['len_header']['end'] = self._io.pos()
            self._debug['magic']['start'] = self._io.pos()
            self.magic = self._io.read_bytes(4)
            self._debug['magic']['end'] = self._io.pos()
            if not self.magic == b"\x4C\x66\x4C\x65":
                raise kaitaistruct.ValidationNotEqualError(b"\x4C\x66\x4C\x65", self.magic, self._io, u"/types/header/seq/1")
            self._debug['version_major']['start'] = self._io.pos()
            self.version_major = self._io.read_u4le()
            self._debug['version_major']['end'] = self._io.pos()
            self._debug['version_minor']['start'] = self._io.pos()
            self.version_minor = self._io.read_u4le()
            self._debug['version_minor']['end'] = self._io.pos()
            self._debug['ofs_start']['start'] = self._io.pos()
            self.ofs_start = self._io.read_u4le()
            self._debug['ofs_start']['end'] = self._io.pos()
            self._debug['ofs_end']['start'] = self._io.pos()
            self.ofs_end = self._io.read_u4le()
            self._debug['ofs_end']['end'] = self._io.pos()
            self._debug['cur_rec_idx']['start'] = self._io.pos()
            self.cur_rec_idx = self._io.read_u4le()
            self._debug['cur_rec_idx']['end'] = self._io.pos()
            self._debug['oldest_rec_idx']['start'] = self._io.pos()
            self.oldest_rec_idx = self._io.read_u4le()
            self._debug['oldest_rec_idx']['end'] = self._io.pos()
            self._debug['len_file_max']['start'] = self._io.pos()
            self.len_file_max = self._io.read_u4le()
            self._debug['len_file_max']['end'] = self._io.pos()
            self._debug['flags']['start'] = self._io.pos()
            self.flags = WindowsEvtLog.Header.Flags(self._io, self, self._root)
            self.flags._read()
            self._debug['flags']['end'] = self._io.pos()
            self._debug['retention']['start'] = self._io.pos()
            self.retention = self._io.read_u4le()
            self._debug['retention']['end'] = self._io.pos()
            self._debug['len_header_2']['start'] = self._io.pos()
            self.len_header_2 = self._io.read_u4le()
            self._debug['len_header_2']['end'] = self._io.pos()

        class Flags(KaitaiStruct):
            SEQ_FIELDS = ["reserved", "archive", "log_full", "wrap", "dirty"]
            def __init__(self, _io, _parent=None, _root=None):
                self._io = _io
                self._parent = _parent
                self._root = _root if _root else self
                self._debug = collections.defaultdict(dict)

            def _read(self):
                self._debug['reserved']['start'] = self._io.pos()
                self.reserved = self._io.read_bits_int_be(28)
                self._debug['reserved']['end'] = self._io.pos()
                self._debug['archive']['start'] = self._io.pos()
                self.archive = self._io.read_bits_int_be(1) != 0
                self._debug['archive']['end'] = self._io.pos()
                self._debug['log_full']['start'] = self._io.pos()
                self.log_full = self._io.read_bits_int_be(1) != 0
                self._debug['log_full']['end'] = self._io.pos()
                self._debug['wrap']['start'] = self._io.pos()
                self.wrap = self._io.read_bits_int_be(1) != 0
                self._debug['wrap']['end'] = self._io.pos()
                self._debug['dirty']['start'] = self._io.pos()
                self.dirty = self._io.read_bits_int_be(1) != 0
                self._debug['dirty']['end'] = self._io.pos()



    class Record(KaitaiStruct):
        """
        .. seealso::
           Source - https://docs.microsoft.com/en-us/windows/win32/api/winnt/ns-winnt-eventlogrecord
        """
        SEQ_FIELDS = ["len_record", "type", "body", "len_record2"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['len_record']['start'] = self._io.pos()
            self.len_record = self._io.read_u4le()
            self._debug['len_record']['end'] = self._io.pos()
            self._debug['type']['start'] = self._io.pos()
            self.type = self._io.read_u4le()
            self._debug['type']['end'] = self._io.pos()
            self._debug['body']['start'] = self._io.pos()
            _on = self.type
            if _on == 1699505740:
                self._raw_body = self._io.read_bytes((self.len_record - 12))
                _io__raw_body = KaitaiStream(BytesIO(self._raw_body))
                self.body = WindowsEvtLog.RecordBody(_io__raw_body, self, self._root)
                self.body._read()
            elif _on == 286331153:
                self._raw_body = self._io.read_bytes((self.len_record - 12))
                _io__raw_body = KaitaiStream(BytesIO(self._raw_body))
                self.body = WindowsEvtLog.CursorRecordBody(_io__raw_body, self, self._root)
                self.body._read()
            else:
                self.body = self._io.read_bytes((self.len_record - 12))
            self._debug['body']['end'] = self._io.pos()
            self._debug['len_record2']['start'] = self._io.pos()
            self.len_record2 = self._io.read_u4le()
            self._debug['len_record2']['end'] = self._io.pos()


    class RecordBody(KaitaiStruct):
        """
        .. seealso::
           Source - https://docs.microsoft.com/en-us/windows/win32/api/winnt/ns-winnt-eventlogrecord
        """

        class EventTypes(Enum):
            error = 1
            audit_failure = 2
            audit_success = 3
            info = 4
            warning = 5
        SEQ_FIELDS = ["idx", "time_generated", "time_written", "event_id", "event_type", "num_strings", "event_category", "reserved", "ofs_strings", "len_user_sid", "ofs_user_sid", "len_data", "ofs_data"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['idx']['start'] = self._io.pos()
            self.idx = self._io.read_u4le()
            self._debug['idx']['end'] = self._io.pos()
            self._debug['time_generated']['start'] = self._io.pos()
            self.time_generated = self._io.read_u4le()
            self._debug['time_generated']['end'] = self._io.pos()
            self._debug['time_written']['start'] = self._io.pos()
            self.time_written = self._io.read_u4le()
            self._debug['time_written']['end'] = self._io.pos()
            self._debug['event_id']['start'] = self._io.pos()
            self.event_id = self._io.read_u4le()
            self._debug['event_id']['end'] = self._io.pos()
            self._debug['event_type']['start'] = self._io.pos()
            self.event_type = KaitaiStream.resolve_enum(WindowsEvtLog.RecordBody.EventTypes, self._io.read_u2le())
            self._debug['event_type']['end'] = self._io.pos()
            self._debug['num_strings']['start'] = self._io.pos()
            self.num_strings = self._io.read_u2le()
            self._debug['num_strings']['end'] = self._io.pos()
            self._debug['event_category']['start'] = self._io.pos()
            self.event_category = self._io.read_u2le()
            self._debug['event_category']['end'] = self._io.pos()
            self._debug['reserved']['start'] = self._io.pos()
            self.reserved = self._io.read_bytes(6)
            self._debug['reserved']['end'] = self._io.pos()
            self._debug['ofs_strings']['start'] = self._io.pos()
            self.ofs_strings = self._io.read_u4le()
            self._debug['ofs_strings']['end'] = self._io.pos()
            self._debug['len_user_sid']['start'] = self._io.pos()
            self.len_user_sid = self._io.read_u4le()
            self._debug['len_user_sid']['end'] = self._io.pos()
            self._debug['ofs_user_sid']['start'] = self._io.pos()
            self.ofs_user_sid = self._io.read_u4le()
            self._debug['ofs_user_sid']['end'] = self._io.pos()
            self._debug['len_data']['start'] = self._io.pos()
            self.len_data = self._io.read_u4le()
            self._debug['len_data']['end'] = self._io.pos()
            self._debug['ofs_data']['start'] = self._io.pos()
            self.ofs_data = self._io.read_u4le()
            self._debug['ofs_data']['end'] = self._io.pos()

        @property
        def user_sid(self):
            if hasattr(self, '_m_user_sid'):
                return self._m_user_sid if hasattr(self, '_m_user_sid') else None

            _pos = self._io.pos()
            self._io.seek((self.ofs_user_sid - 8))
            self._debug['_m_user_sid']['start'] = self._io.pos()
            self._m_user_sid = self._io.read_bytes(self.len_user_sid)
            self._debug['_m_user_sid']['end'] = self._io.pos()
            self._io.seek(_pos)
            return self._m_user_sid if hasattr(self, '_m_user_sid') else None

        @property
        def data(self):
            if hasattr(self, '_m_data'):
                return self._m_data if hasattr(self, '_m_data') else None

            _pos = self._io.pos()
            self._io.seek((self.ofs_data - 8))
            self._debug['_m_data']['start'] = self._io.pos()
            self._m_data = self._io.read_bytes(self.len_data)
            self._debug['_m_data']['end'] = self._io.pos()
            self._io.seek(_pos)
            return self._m_data if hasattr(self, '_m_data') else None


    class CursorRecordBody(KaitaiStruct):
        """
        .. seealso::
           Source - http://www.forensicswiki.xyz/page/Windows_Event_Log_(EVT)#Cursor_Record
        """
        SEQ_FIELDS = ["magic", "ofs_first_record", "ofs_next_record", "idx_next_record", "idx_first_record"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['magic']['start'] = self._io.pos()
            self.magic = self._io.read_bytes(12)
            self._debug['magic']['end'] = self._io.pos()
            if not self.magic == b"\x22\x22\x22\x22\x33\x33\x33\x33\x44\x44\x44\x44":
                raise kaitaistruct.ValidationNotEqualError(b"\x22\x22\x22\x22\x33\x33\x33\x33\x44\x44\x44\x44", self.magic, self._io, u"/types/cursor_record_body/seq/0")
            self._debug['ofs_first_record']['start'] = self._io.pos()
            self.ofs_first_record = self._io.read_u4le()
            self._debug['ofs_first_record']['end'] = self._io.pos()
            self._debug['ofs_next_record']['start'] = self._io.pos()
            self.ofs_next_record = self._io.read_u4le()
            self._debug['ofs_next_record']['end'] = self._io.pos()
            self._debug['idx_next_record']['start'] = self._io.pos()
            self.idx_next_record = self._io.read_u4le()
            self._debug['idx_next_record']['end'] = self._io.pos()
            self._debug['idx_first_record']['start'] = self._io.pos()
            self.idx_first_record = self._io.read_u4le()
            self._debug['idx_first_record']['end'] = self._io.pos()






# This is a generated file! Please edit source .ksy file and use kaitai-struct-compiler to rebuild

from pkg_resources import parse_version
import kaitaistruct
from kaitaistruct import KaitaiStruct, KaitaiStream, BytesIO
import collections


if parse_version(kaitaistruct.__version__) < parse_version('0.9'):
    raise Exception("Incompatible Kaitai Struct Python API: 0.9 or later is required, but you have %s" % (kaitaistruct.__version__))

class WindowsSystemtime(KaitaiStruct):
    """Microsoft Windows SYSTEMTIME structure, stores individual components
    of date and time as individual fields, up to millisecond precision.
    
    .. seealso::
       Source - https://docs.microsoft.com/en-us/windows/win32/api/minwinbase/ns-minwinbase-systemtime
    """
    SEQ_FIELDS = ["year", "month", "dow", "day", "hour", "min", "sec", "msec"]
    def __init__(self, _io, _parent=None, _root=None):
        self._io = _io
        self._parent = _parent
        self._root = _root if _root else self
        self._debug = collections.defaultdict(dict)

    def _read(self):
        self._debug['year']['start'] = self._io.pos()
        self.year = self._io.read_u2le()
        self._debug['year']['end'] = self._io.pos()
        self._debug['month']['start'] = self._io.pos()
        self.month = self._io.read_u2le()
        self._debug['month']['end'] = self._io.pos()
        self._debug['dow']['start'] = self._io.pos()
        self.dow = self._io.read_u2le()
        self._debug['dow']['end'] = self._io.pos()
        self._debug['day']['start'] = self._io.pos()
        self.day = self._io.read_u2le()
        self._debug['day']['end'] = self._io.pos()
        self._debug['hour']['start'] = self._io.pos()
        self.hour = self._io.read_u2le()
        self._debug['hour']['end'] = self._io.pos()
        self._debug['min']['start'] = self._io.pos()
        self.min = self._io.read_u2le()
        self._debug['min']['end'] = self._io.pos()
        self._debug['sec']['start'] = self._io.pos()
        self.sec = self._io.read_u2le()
        self._debug['sec']['end'] = self._io.pos()
        self._debug['msec']['start'] = self._io.pos()
        self.msec = self._io.read_u2le()
        self._debug['msec']['end'] = self._io.pos()





# This is a generated file! Please edit source .ksy file and use kaitai-struct-compiler to rebuild

from pkg_resources import parse_version
import kaitaistruct
from kaitaistruct import KaitaiStruct, KaitaiStream, BytesIO
import collections
from enum import Enum


if parse_version(kaitaistruct.__version__) < parse_version('0.9'):
    raise Exception("Incompatible Kaitai Struct Python API: 0.9 or later is required, but you have %s" % (kaitaistruct.__version__))

class Cramfs(KaitaiStruct):
    SEQ_FIELDS = ["super_block"]
    def __init__(self, _io, _parent=None, _root=None):
        self._io = _io
        self._parent = _parent
        self._root = _root if _root else self
        self._debug = collections.defaultdict(dict)

    def _read(self):
        self._debug['super_block']['start'] = self._io.pos()
        self.super_block = Cramfs.SuperBlockStruct(self._io, self, self._root)
        self.super_block._read()
        self._debug['super_block']['end'] = self._io.pos()

    class SuperBlockStruct(KaitaiStruct):
        SEQ_FIELDS = ["magic", "size", "flags", "future", "signature", "fsid", "name", "root"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['magic']['start'] = self._io.pos()
            self.magic = self._io.read_bytes(4)
            self._debug['magic']['end'] = self._io.pos()
            if not self.magic == b"\x45\x3D\xCD\x28":
                raise kaitaistruct.ValidationNotEqualError(b"\x45\x3D\xCD\x28", self.magic, self._io, u"/types/super_block_struct/seq/0")
            self._debug['size']['start'] = self._io.pos()
            self.size = self._io.read_u4le()
            self._debug['size']['end'] = self._io.pos()
            self._debug['flags']['start'] = self._io.pos()
            self.flags = self._io.read_u4le()
            self._debug['flags']['end'] = self._io.pos()
            self._debug['future']['start'] = self._io.pos()
            self.future = self._io.read_u4le()
            self._debug['future']['end'] = self._io.pos()
            self._debug['signature']['start'] = self._io.pos()
            self.signature = self._io.read_bytes(16)
            self._debug['signature']['end'] = self._io.pos()
            if not self.signature == b"\x43\x6F\x6D\x70\x72\x65\x73\x73\x65\x64\x20\x52\x4F\x4D\x46\x53":
                raise kaitaistruct.ValidationNotEqualError(b"\x43\x6F\x6D\x70\x72\x65\x73\x73\x65\x64\x20\x52\x4F\x4D\x46\x53", self.signature, self._io, u"/types/super_block_struct/seq/4")
            self._debug['fsid']['start'] = self._io.pos()
            self.fsid = Cramfs.Info(self._io, self, self._root)
            self.fsid._read()
            self._debug['fsid']['end'] = self._io.pos()
            self._debug['name']['start'] = self._io.pos()
            self.name = (self._io.read_bytes(16)).decode(u"ASCII")
            self._debug['name']['end'] = self._io.pos()
            self._debug['root']['start'] = self._io.pos()
            self.root = Cramfs.Inode(self._io, self, self._root)
            self.root._read()
            self._debug['root']['end'] = self._io.pos()

        @property
        def flag_fsid_v2(self):
            if hasattr(self, '_m_flag_fsid_v2'):
                return self._m_flag_fsid_v2 if hasattr(self, '_m_flag_fsid_v2') else None

            self._m_flag_fsid_v2 = ((self.flags >> 0) & 1)
            return self._m_flag_fsid_v2 if hasattr(self, '_m_flag_fsid_v2') else None

        @property
        def flag_holes(self):
            if hasattr(self, '_m_flag_holes'):
                return self._m_flag_holes if hasattr(self, '_m_flag_holes') else None

            self._m_flag_holes = ((self.flags >> 8) & 1)
            return self._m_flag_holes if hasattr(self, '_m_flag_holes') else None

        @property
        def flag_wrong_signature(self):
            if hasattr(self, '_m_flag_wrong_signature'):
                return self._m_flag_wrong_signature if hasattr(self, '_m_flag_wrong_signature') else None

            self._m_flag_wrong_signature = ((self.flags >> 9) & 1)
            return self._m_flag_wrong_signature if hasattr(self, '_m_flag_wrong_signature') else None

        @property
        def flag_sorted_dirs(self):
            if hasattr(self, '_m_flag_sorted_dirs'):
                return self._m_flag_sorted_dirs if hasattr(self, '_m_flag_sorted_dirs') else None

            self._m_flag_sorted_dirs = ((self.flags >> 1) & 1)
            return self._m_flag_sorted_dirs if hasattr(self, '_m_flag_sorted_dirs') else None

        @property
        def flag_shifted_root_offset(self):
            if hasattr(self, '_m_flag_shifted_root_offset'):
                return self._m_flag_shifted_root_offset if hasattr(self, '_m_flag_shifted_root_offset') else None

            self._m_flag_shifted_root_offset = ((self.flags >> 10) & 1)
            return self._m_flag_shifted_root_offset if hasattr(self, '_m_flag_shifted_root_offset') else None


    class ChunkedDataInode(KaitaiStruct):
        SEQ_FIELDS = ["block_end_index", "raw_blocks"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['block_end_index']['start'] = self._io.pos()
            self.block_end_index = [None] * (((self._parent.size + self._root.page_size) - 1) // self._root.page_size)
            for i in range(((self._parent.size + self._root.page_size) - 1) // self._root.page_size):
                if not 'arr' in self._debug['block_end_index']:
                    self._debug['block_end_index']['arr'] = []
                self._debug['block_end_index']['arr'].append({'start': self._io.pos()})
                self.block_end_index[i] = self._io.read_u4le()
                self._debug['block_end_index']['arr'][i]['end'] = self._io.pos()

            self._debug['block_end_index']['end'] = self._io.pos()
            self._debug['raw_blocks']['start'] = self._io.pos()
            self.raw_blocks = self._io.read_bytes_full()
            self._debug['raw_blocks']['end'] = self._io.pos()


    class Inode(KaitaiStruct):

        class FileType(Enum):
            fifo = 1
            chrdev = 2
            dir = 4
            blkdev = 6
            reg_file = 8
            symlink = 10
            socket = 12
        SEQ_FIELDS = ["mode", "uid", "size_gid", "namelen_offset", "name"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['mode']['start'] = self._io.pos()
            self.mode = self._io.read_u2le()
            self._debug['mode']['end'] = self._io.pos()
            self._debug['uid']['start'] = self._io.pos()
            self.uid = self._io.read_u2le()
            self._debug['uid']['end'] = self._io.pos()
            self._debug['size_gid']['start'] = self._io.pos()
            self.size_gid = self._io.read_u4le()
            self._debug['size_gid']['end'] = self._io.pos()
            self._debug['namelen_offset']['start'] = self._io.pos()
            self.namelen_offset = self._io.read_u4le()
            self._debug['namelen_offset']['end'] = self._io.pos()
            self._debug['name']['start'] = self._io.pos()
            self.name = (self._io.read_bytes(self.namelen)).decode(u"utf-8")
            self._debug['name']['end'] = self._io.pos()

        @property
        def attr(self):
            if hasattr(self, '_m_attr'):
                return self._m_attr if hasattr(self, '_m_attr') else None

            self._m_attr = ((self.mode >> 9) & 7)
            return self._m_attr if hasattr(self, '_m_attr') else None

        @property
        def as_reg_file(self):
            if hasattr(self, '_m_as_reg_file'):
                return self._m_as_reg_file if hasattr(self, '_m_as_reg_file') else None

            io = self._root._io
            _pos = io.pos()
            io.seek(self.offset)
            self._debug['_m_as_reg_file']['start'] = io.pos()
            self._m_as_reg_file = Cramfs.ChunkedDataInode(io, self, self._root)
            self._m_as_reg_file._read()
            self._debug['_m_as_reg_file']['end'] = io.pos()
            io.seek(_pos)
            return self._m_as_reg_file if hasattr(self, '_m_as_reg_file') else None

        @property
        def perm_u(self):
            if hasattr(self, '_m_perm_u'):
                return self._m_perm_u if hasattr(self, '_m_perm_u') else None

            self._m_perm_u = ((self.mode >> 6) & 7)
            return self._m_perm_u if hasattr(self, '_m_perm_u') else None

        @property
        def as_symlink(self):
            if hasattr(self, '_m_as_symlink'):
                return self._m_as_symlink if hasattr(self, '_m_as_symlink') else None

            io = self._root._io
            _pos = io.pos()
            io.seek(self.offset)
            self._debug['_m_as_symlink']['start'] = io.pos()
            self._m_as_symlink = Cramfs.ChunkedDataInode(io, self, self._root)
            self._m_as_symlink._read()
            self._debug['_m_as_symlink']['end'] = io.pos()
            io.seek(_pos)
            return self._m_as_symlink if hasattr(self, '_m_as_symlink') else None

        @property
        def perm_o(self):
            if hasattr(self, '_m_perm_o'):
                return self._m_perm_o if hasattr(self, '_m_perm_o') else None

            self._m_perm_o = (self.mode & 7)
            return self._m_perm_o if hasattr(self, '_m_perm_o') else None

        @property
        def size(self):
            if hasattr(self, '_m_size'):
                return self._m_size if hasattr(self, '_m_size') else None

            self._m_size = (self.size_gid & 16777215)
            return self._m_size if hasattr(self, '_m_size') else None

        @property
        def gid(self):
            if hasattr(self, '_m_gid'):
                return self._m_gid if hasattr(self, '_m_gid') else None

            self._m_gid = (self.size_gid >> 24)
            return self._m_gid if hasattr(self, '_m_gid') else None

        @property
        def perm_g(self):
            if hasattr(self, '_m_perm_g'):
                return self._m_perm_g if hasattr(self, '_m_perm_g') else None

            self._m_perm_g = ((self.mode >> 3) & 7)
            return self._m_perm_g if hasattr(self, '_m_perm_g') else None

        @property
        def namelen(self):
            if hasattr(self, '_m_namelen'):
                return self._m_namelen if hasattr(self, '_m_namelen') else None

            self._m_namelen = ((self.namelen_offset & 63) << 2)
            return self._m_namelen if hasattr(self, '_m_namelen') else None

        @property
        def as_dir(self):
            if hasattr(self, '_m_as_dir'):
                return self._m_as_dir if hasattr(self, '_m_as_dir') else None

            io = self._root._io
            _pos = io.pos()
            io.seek(self.offset)
            self._debug['_m_as_dir']['start'] = io.pos()
            self._raw__m_as_dir = io.read_bytes(self.size)
            _io__raw__m_as_dir = KaitaiStream(BytesIO(self._raw__m_as_dir))
            self._m_as_dir = Cramfs.DirInode(_io__raw__m_as_dir, self, self._root)
            self._m_as_dir._read()
            self._debug['_m_as_dir']['end'] = io.pos()
            io.seek(_pos)
            return self._m_as_dir if hasattr(self, '_m_as_dir') else None

        @property
        def type(self):
            if hasattr(self, '_m_type'):
                return self._m_type if hasattr(self, '_m_type') else None

            self._m_type = KaitaiStream.resolve_enum(Cramfs.Inode.FileType, ((self.mode >> 12) & 15))
            return self._m_type if hasattr(self, '_m_type') else None

        @property
        def offset(self):
            if hasattr(self, '_m_offset'):
                return self._m_offset if hasattr(self, '_m_offset') else None

            self._m_offset = (((self.namelen_offset >> 6) & 67108863) << 2)
            return self._m_offset if hasattr(self, '_m_offset') else None


    class DirInode(KaitaiStruct):
        SEQ_FIELDS = ["children"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            if self._io.size() > 0:
                self._debug['children']['start'] = self._io.pos()
                self.children = []
                i = 0
                while not self._io.is_eof():
                    if not 'arr' in self._debug['children']:
                        self._debug['children']['arr'] = []
                    self._debug['children']['arr'].append({'start': self._io.pos()})
                    _t_children = Cramfs.Inode(self._io, self, self._root)
                    _t_children._read()
                    self.children.append(_t_children)
                    self._debug['children']['arr'][len(self.children) - 1]['end'] = self._io.pos()
                    i += 1

                self._debug['children']['end'] = self._io.pos()



    class Info(KaitaiStruct):
        SEQ_FIELDS = ["crc", "edition", "blocks", "files"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['crc']['start'] = self._io.pos()
            self.crc = self._io.read_u4le()
            self._debug['crc']['end'] = self._io.pos()
            self._debug['edition']['start'] = self._io.pos()
            self.edition = self._io.read_u4le()
            self._debug['edition']['end'] = self._io.pos()
            self._debug['blocks']['start'] = self._io.pos()
            self.blocks = self._io.read_u4le()
            self._debug['blocks']['end'] = self._io.pos()
            self._debug['files']['start'] = self._io.pos()
            self.files = self._io.read_u4le()
            self._debug['files']['end'] = self._io.pos()


    @property
    def page_size(self):
        if hasattr(self, '_m_page_size'):
            return self._m_page_size if hasattr(self, '_m_page_size') else None

        self._m_page_size = 4096
        return self._m_page_size if hasattr(self, '_m_page_size') else None





# This is a generated file! Please edit source .ksy file and use kaitai-struct-compiler to rebuild

from pkg_resources import parse_version
import kaitaistruct
from kaitaistruct import KaitaiStruct, KaitaiStream, BytesIO
import collections


if parse_version(kaitaistruct.__version__) < parse_version('0.9'):
    raise Exception("Incompatible Kaitai Struct Python API: 0.9 or later is required, but you have %s" % (kaitaistruct.__version__))

class Utf8String(KaitaiStruct):
    """UTF-8 is a popular character encoding scheme that allows to
    represent strings as sequence of code points defined in Unicode
    standard. Its features are:
    
    * variable width (i.e. one code point might be represented by 1 to 4
      bytes)
    * backward compatiblity with ASCII
    * basic validity checking (and thus distinguishing from other legacy
      8-bit encodings)
    * maintaining sort order of codepoints if sorted as a byte array
    
    WARNING: For the vast majority of practical purposes of format
    definitions in Kaitai Struct, you'd likely NOT want to use this and
    rather just use `type: str` with `encoding: utf-8`. That will use
    native string implementations, which are most likely more efficient
    and will give you native language strings, rather than an array of
    individual codepoints.  This format definition is provided mostly
    for educational / research purposes.
    """
    SEQ_FIELDS = ["codepoints"]
    def __init__(self, _io, _parent=None, _root=None):
        self._io = _io
        self._parent = _parent
        self._root = _root if _root else self
        self._debug = collections.defaultdict(dict)

    def _read(self):
        self._debug['codepoints']['start'] = self._io.pos()
        self.codepoints = []
        i = 0
        while not self._io.is_eof():
            if not 'arr' in self._debug['codepoints']:
                self._debug['codepoints']['arr'] = []
            self._debug['codepoints']['arr'].append({'start': self._io.pos()})
            _t_codepoints = Utf8String.Utf8Codepoint(self._io.pos(), self._io, self, self._root)
            _t_codepoints._read()
            self.codepoints.append(_t_codepoints)
            self._debug['codepoints']['arr'][len(self.codepoints) - 1]['end'] = self._io.pos()
            i += 1

        self._debug['codepoints']['end'] = self._io.pos()

    class Utf8Codepoint(KaitaiStruct):
        SEQ_FIELDS = ["bytes"]
        def __init__(self, ofs, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self.ofs = ofs
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['bytes']['start'] = self._io.pos()
            self.bytes = self._io.read_bytes(self.len_bytes)
            self._debug['bytes']['end'] = self._io.pos()

        @property
        def raw1(self):
            if hasattr(self, '_m_raw1'):
                return self._m_raw1 if hasattr(self, '_m_raw1') else None

            if self.len_bytes >= 2:
                self._m_raw1 = (KaitaiStream.byte_array_index(self.bytes, 1) & 63)

            return self._m_raw1 if hasattr(self, '_m_raw1') else None

        @property
        def len_bytes(self):
            if hasattr(self, '_m_len_bytes'):
                return self._m_len_bytes if hasattr(self, '_m_len_bytes') else None

            self._m_len_bytes = (1 if (self.byte0 & 128) == 0 else (2 if (self.byte0 & 224) == 192 else (3 if (self.byte0 & 240) == 224 else (4 if (self.byte0 & 248) == 240 else -1))))
            return self._m_len_bytes if hasattr(self, '_m_len_bytes') else None

        @property
        def raw3(self):
            if hasattr(self, '_m_raw3'):
                return self._m_raw3 if hasattr(self, '_m_raw3') else None

            if self.len_bytes >= 4:
                self._m_raw3 = (KaitaiStream.byte_array_index(self.bytes, 3) & 63)

            return self._m_raw3 if hasattr(self, '_m_raw3') else None

        @property
        def value_as_int(self):
            if hasattr(self, '_m_value_as_int'):
                return self._m_value_as_int if hasattr(self, '_m_value_as_int') else None

            self._m_value_as_int = (self.raw0 if self.len_bytes == 1 else (((self.raw0 << 6) | self.raw1) if self.len_bytes == 2 else ((((self.raw0 << 12) | (self.raw1 << 6)) | self.raw2) if self.len_bytes == 3 else (((((self.raw0 << 18) | (self.raw1 << 12)) | (self.raw2 << 6)) | self.raw3) if self.len_bytes == 4 else -1))))
            return self._m_value_as_int if hasattr(self, '_m_value_as_int') else None

        @property
        def raw0(self):
            if hasattr(self, '_m_raw0'):
                return self._m_raw0 if hasattr(self, '_m_raw0') else None

            self._m_raw0 = (KaitaiStream.byte_array_index(self.bytes, 0) & (127 if self.len_bytes == 1 else (31 if self.len_bytes == 2 else (15 if self.len_bytes == 3 else (7 if self.len_bytes == 4 else 0)))))
            return self._m_raw0 if hasattr(self, '_m_raw0') else None

        @property
        def byte0(self):
            if hasattr(self, '_m_byte0'):
                return self._m_byte0 if hasattr(self, '_m_byte0') else None

            _pos = self._io.pos()
            self._io.seek(self.ofs)
            self._debug['_m_byte0']['start'] = self._io.pos()
            self._m_byte0 = self._io.read_u1()
            self._debug['_m_byte0']['end'] = self._io.pos()
            self._io.seek(_pos)
            return self._m_byte0 if hasattr(self, '_m_byte0') else None

        @property
        def raw2(self):
            if hasattr(self, '_m_raw2'):
                return self._m_raw2 if hasattr(self, '_m_raw2') else None

            if self.len_bytes >= 3:
                self._m_raw2 = (KaitaiStream.byte_array_index(self.bytes, 2) & 63)

            return self._m_raw2 if hasattr(self, '_m_raw2') else None






# This is a generated file! Please edit source .ksy file and use kaitai-struct-compiler to rebuild

from pkg_resources import parse_version
import kaitaistruct
from kaitaistruct import KaitaiStruct, KaitaiStream, BytesIO
import collections


if parse_version(kaitaistruct.__version__) < parse_version('0.9'):
    raise Exception("Incompatible Kaitai Struct Python API: 0.9 or later is required, but you have %s" % (kaitaistruct.__version__))

class AndroidOpenglShadersCache(KaitaiStruct):
    """Android apps using directly or indirectly OpenGL cache compiled shaders
    into com.android.opengl.shaders_cache file.
    
    .. seealso::
       Source - https://android.googlesource.com/platform/frameworks/native/+/master/opengl/libs/EGL/FileBlobCache.cpp
    """
    SEQ_FIELDS = ["magic", "crc32", "contents"]
    def __init__(self, _io, _parent=None, _root=None):
        self._io = _io
        self._parent = _parent
        self._root = _root if _root else self
        self._debug = collections.defaultdict(dict)

    def _read(self):
        self._debug['magic']['start'] = self._io.pos()
        self.magic = self._io.read_bytes(4)
        self._debug['magic']['end'] = self._io.pos()
        if not self.magic == b"\x45\x47\x4C\x24":
            raise kaitaistruct.ValidationNotEqualError(b"\x45\x47\x4C\x24", self.magic, self._io, u"/seq/0")
        self._debug['crc32']['start'] = self._io.pos()
        self.crc32 = self._io.read_u4le()
        self._debug['crc32']['end'] = self._io.pos()
        self._debug['contents']['start'] = self._io.pos()
        self._raw_contents = self._io.read_bytes_full()
        _io__raw_contents = KaitaiStream(BytesIO(self._raw_contents))
        self.contents = AndroidOpenglShadersCache.Cache(_io__raw_contents, self, self._root)
        self.contents._read()
        self._debug['contents']['end'] = self._io.pos()

    class Alignment(KaitaiStruct):
        SEQ_FIELDS = ["alignment"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['alignment']['start'] = self._io.pos()
            self.alignment = self._io.read_bytes(((self._io.pos() + 3) & (~3 - self._io.pos())))
            self._debug['alignment']['end'] = self._io.pos()


    class PrefixedString(KaitaiStruct):
        SEQ_FIELDS = ["len_str", "str", "alignment"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['len_str']['start'] = self._io.pos()
            self.len_str = self._io.read_u4le()
            self._debug['len_str']['end'] = self._io.pos()
            self._debug['str']['start'] = self._io.pos()
            self.str = (KaitaiStream.bytes_terminate(self._io.read_bytes(self.len_str), 0, False)).decode(u"ascii")
            self._debug['str']['end'] = self._io.pos()
            self._debug['alignment']['start'] = self._io.pos()
            self.alignment = AndroidOpenglShadersCache.Alignment(self._io, self, self._root)
            self.alignment._read()
            self._debug['alignment']['end'] = self._io.pos()


    class Cache(KaitaiStruct):
        """
        .. seealso::
           Source - https://android.googlesource.com/platform/frameworks/native/+/master/opengl/libs/EGL/BlobCache.cpp
        """
        SEQ_FIELDS = ["magic", "version", "device_version", "num_entries", "build_id", "entries"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['magic']['start'] = self._io.pos()
            self.magic = self._io.read_bytes(4)
            self._debug['magic']['end'] = self._io.pos()
            if not self.magic == b"\x24\x62\x42\x5F":
                raise kaitaistruct.ValidationNotEqualError(b"\x24\x62\x42\x5F", self.magic, self._io, u"/types/cache/seq/0")
            self._debug['version']['start'] = self._io.pos()
            self.version = self._io.read_u4le()
            self._debug['version']['end'] = self._io.pos()
            self._debug['device_version']['start'] = self._io.pos()
            self.device_version = self._io.read_u4le()
            self._debug['device_version']['end'] = self._io.pos()
            self._debug['num_entries']['start'] = self._io.pos()
            self.num_entries = self._io.read_u4le()
            self._debug['num_entries']['end'] = self._io.pos()
            if self.version >= 3:
                self._debug['build_id']['start'] = self._io.pos()
                self.build_id = AndroidOpenglShadersCache.PrefixedString(self._io, self, self._root)
                self.build_id._read()
                self._debug['build_id']['end'] = self._io.pos()

            self._debug['entries']['start'] = self._io.pos()
            self.entries = [None] * (self.num_entries)
            for i in range(self.num_entries):
                if not 'arr' in self._debug['entries']:
                    self._debug['entries']['arr'] = []
                self._debug['entries']['arr'].append({'start': self._io.pos()})
                _t_entries = AndroidOpenglShadersCache.Cache.Entry(self._io, self, self._root)
                _t_entries._read()
                self.entries[i] = _t_entries
                self._debug['entries']['arr'][i]['end'] = self._io.pos()

            self._debug['entries']['end'] = self._io.pos()

        class Entry(KaitaiStruct):
            SEQ_FIELDS = ["len_key", "len_value", "key", "value", "alignment"]
            def __init__(self, _io, _parent=None, _root=None):
                self._io = _io
                self._parent = _parent
                self._root = _root if _root else self
                self._debug = collections.defaultdict(dict)

            def _read(self):
                self._debug['len_key']['start'] = self._io.pos()
                self.len_key = self._io.read_u4le()
                self._debug['len_key']['end'] = self._io.pos()
                self._debug['len_value']['start'] = self._io.pos()
                self.len_value = self._io.read_u4le()
                self._debug['len_value']['end'] = self._io.pos()
                self._debug['key']['start'] = self._io.pos()
                self.key = self._io.read_bytes(self.len_key)
                self._debug['key']['end'] = self._io.pos()
                self._debug['value']['start'] = self._io.pos()
                self.value = self._io.read_bytes(self.len_value)
                self._debug['value']['end'] = self._io.pos()
                self._debug['alignment']['start'] = self._io.pos()
                self.alignment = AndroidOpenglShadersCache.Alignment(self._io, self, self._root)
                self.alignment._read()
                self._debug['alignment']['end'] = self._io.pos()







# This is a generated file! Please edit source .ksy file and use kaitai-struct-compiler to rebuild

from pkg_resources import parse_version
import kaitaistruct
from kaitaistruct import KaitaiStruct, KaitaiStream, BytesIO
from enum import Enum
import collections


if parse_version(kaitaistruct.__version__) < parse_version('0.9'):
    raise Exception("Incompatible Kaitai Struct Python API: 0.9 or later is required, but you have %s" % (kaitaistruct.__version__))

class SudoersTs(KaitaiStruct):
    """This spec can be used to parse sudo time stamp files located in directories
    such as /run/sudo/ts/$USER or /var/lib/sudo/ts/$USER.
    
    .. seealso::
       Source - https://www.sudo.ws/man/1.8.27/sudoers_timestamp.man.html
    """

    class TsType(Enum):
        global = 1
        tty = 2
        ppid = 3
        lockexcl = 4
    SEQ_FIELDS = ["records"]
    def __init__(self, _io, _parent=None, _root=None):
        self._io = _io
        self._parent = _parent
        self._root = _root if _root else self
        self._debug = collections.defaultdict(dict)

    def _read(self):
        self._debug['records']['start'] = self._io.pos()
        self.records = []
        i = 0
        while not self._io.is_eof():
            if not 'arr' in self._debug['records']:
                self._debug['records']['arr'] = []
            self._debug['records']['arr'].append({'start': self._io.pos()})
            _t_records = SudoersTs.Record(self._io, self, self._root)
            _t_records._read()
            self.records.append(_t_records)
            self._debug['records']['arr'][len(self.records) - 1]['end'] = self._io.pos()
            i += 1

        self._debug['records']['end'] = self._io.pos()

    class RecordV2(KaitaiStruct):
        SEQ_FIELDS = ["type", "flags", "auth_uid", "sid", "start_time", "ts", "ttydev", "ppid"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['type']['start'] = self._io.pos()
            self.type = KaitaiStream.resolve_enum(SudoersTs.TsType, self._io.read_u2le())
            self._debug['type']['end'] = self._io.pos()
            self._debug['flags']['start'] = self._io.pos()
            self.flags = SudoersTs.TsFlag(self._io, self, self._root)
            self.flags._read()
            self._debug['flags']['end'] = self._io.pos()
            self._debug['auth_uid']['start'] = self._io.pos()
            self.auth_uid = self._io.read_u4le()
            self._debug['auth_uid']['end'] = self._io.pos()
            self._debug['sid']['start'] = self._io.pos()
            self.sid = self._io.read_u4le()
            self._debug['sid']['end'] = self._io.pos()
            self._debug['start_time']['start'] = self._io.pos()
            self.start_time = SudoersTs.Timespec(self._io, self, self._root)
            self.start_time._read()
            self._debug['start_time']['end'] = self._io.pos()
            self._debug['ts']['start'] = self._io.pos()
            self.ts = SudoersTs.Timespec(self._io, self, self._root)
            self.ts._read()
            self._debug['ts']['end'] = self._io.pos()
            if self.type == SudoersTs.TsType.tty:
                self._debug['ttydev']['start'] = self._io.pos()
                self.ttydev = self._io.read_u4le()
                self._debug['ttydev']['end'] = self._io.pos()

            if self.type == SudoersTs.TsType.ppid:
                self._debug['ppid']['start'] = self._io.pos()
                self.ppid = self._io.read_u4le()
                self._debug['ppid']['end'] = self._io.pos()



    class TsFlag(KaitaiStruct):
        SEQ_FIELDS = ["reserved0", "anyuid", "disabled", "reserved1"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['reserved0']['start'] = self._io.pos()
            self.reserved0 = self._io.read_bits_int_be(6)
            self._debug['reserved0']['end'] = self._io.pos()
            self._debug['anyuid']['start'] = self._io.pos()
            self.anyuid = self._io.read_bits_int_be(1) != 0
            self._debug['anyuid']['end'] = self._io.pos()
            self._debug['disabled']['start'] = self._io.pos()
            self.disabled = self._io.read_bits_int_be(1) != 0
            self._debug['disabled']['end'] = self._io.pos()
            self._debug['reserved1']['start'] = self._io.pos()
            self.reserved1 = self._io.read_bits_int_be(8)
            self._debug['reserved1']['end'] = self._io.pos()


    class RecordV1(KaitaiStruct):
        SEQ_FIELDS = ["type", "flags", "auth_uid", "sid", "ts", "ttydev", "ppid"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['type']['start'] = self._io.pos()
            self.type = KaitaiStream.resolve_enum(SudoersTs.TsType, self._io.read_u2le())
            self._debug['type']['end'] = self._io.pos()
            self._debug['flags']['start'] = self._io.pos()
            self.flags = SudoersTs.TsFlag(self._io, self, self._root)
            self.flags._read()
            self._debug['flags']['end'] = self._io.pos()
            self._debug['auth_uid']['start'] = self._io.pos()
            self.auth_uid = self._io.read_u4le()
            self._debug['auth_uid']['end'] = self._io.pos()
            self._debug['sid']['start'] = self._io.pos()
            self.sid = self._io.read_u4le()
            self._debug['sid']['end'] = self._io.pos()
            self._debug['ts']['start'] = self._io.pos()
            self.ts = SudoersTs.Timespec(self._io, self, self._root)
            self.ts._read()
            self._debug['ts']['end'] = self._io.pos()
            if self.type == SudoersTs.TsType.tty:
                self._debug['ttydev']['start'] = self._io.pos()
                self.ttydev = self._io.read_u4le()
                self._debug['ttydev']['end'] = self._io.pos()

            if self.type == SudoersTs.TsType.ppid:
                self._debug['ppid']['start'] = self._io.pos()
                self.ppid = self._io.read_u4le()
                self._debug['ppid']['end'] = self._io.pos()



    class Timespec(KaitaiStruct):
        SEQ_FIELDS = ["sec", "nsec"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['sec']['start'] = self._io.pos()
            self.sec = self._io.read_s8le()
            self._debug['sec']['end'] = self._io.pos()
            self._debug['nsec']['start'] = self._io.pos()
            self.nsec = self._io.read_s8le()
            self._debug['nsec']['end'] = self._io.pos()


    class Record(KaitaiStruct):
        SEQ_FIELDS = ["version", "len_record", "payload"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['version']['start'] = self._io.pos()
            self.version = self._io.read_u2le()
            self._debug['version']['end'] = self._io.pos()
            self._debug['len_record']['start'] = self._io.pos()
            self.len_record = self._io.read_u2le()
            self._debug['len_record']['end'] = self._io.pos()
            self._debug['payload']['start'] = self._io.pos()
            _on = self.version
            if _on == 1:
                self._raw_payload = self._io.read_bytes((self.len_record - 4))
                _io__raw_payload = KaitaiStream(BytesIO(self._raw_payload))
                self.payload = SudoersTs.RecordV1(_io__raw_payload, self, self._root)
                self.payload._read()
            elif _on == 2:
                self._raw_payload = self._io.read_bytes((self.len_record - 4))
                _io__raw_payload = KaitaiStream(BytesIO(self._raw_payload))
                self.payload = SudoersTs.RecordV2(_io__raw_payload, self, self._root)
                self.payload._read()
            else:
                self.payload = self._io.read_bytes((self.len_record - 4))
            self._debug['payload']['end'] = self._io.pos()






# This is a generated file! Please edit source .ksy file and use kaitai-struct-compiler to rebuild

from pkg_resources import parse_version
import kaitaistruct
from kaitaistruct import KaitaiStruct, KaitaiStream, BytesIO
import collections


if parse_version(kaitaistruct.__version__) < parse_version('0.9'):
    raise Exception("Incompatible Kaitai Struct Python API: 0.9 or later is required, but you have %s" % (kaitaistruct.__version__))

class AndesFirmware(KaitaiStruct):
    """Firmware image found with MediaTek MT76xx wifi chipsets."""
    SEQ_FIELDS = ["image_header", "ilm", "dlm"]
    def __init__(self, _io, _parent=None, _root=None):
        self._io = _io
        self._parent = _parent
        self._root = _root if _root else self
        self._debug = collections.defaultdict(dict)

    def _read(self):
        self._debug['image_header']['start'] = self._io.pos()
        self._raw_image_header = self._io.read_bytes(32)
        _io__raw_image_header = KaitaiStream(BytesIO(self._raw_image_header))
        self.image_header = AndesFirmware.ImageHeader(_io__raw_image_header, self, self._root)
        self.image_header._read()
        self._debug['image_header']['end'] = self._io.pos()
        self._debug['ilm']['start'] = self._io.pos()
        self.ilm = self._io.read_bytes(self.image_header.ilm_len)
        self._debug['ilm']['end'] = self._io.pos()
        self._debug['dlm']['start'] = self._io.pos()
        self.dlm = self._io.read_bytes(self.image_header.dlm_len)
        self._debug['dlm']['end'] = self._io.pos()

    class ImageHeader(KaitaiStruct):
        SEQ_FIELDS = ["ilm_len", "dlm_len", "fw_ver", "build_ver", "extra", "build_time"]
        def __init__(self, _io, _parent=None, _root=None):
            self._io = _io
            self._parent = _parent
            self._root = _root if _root else self
            self._debug = collections.defaultdict(dict)

        def _read(self):
            self._debug['ilm_len']['start'] = self._io.pos()
            self.ilm_len = self._io.read_u4le()
            self._debug['ilm_len']['end'] = self._io.pos()
            self._debug['dlm_len']['start'] = self._io.pos()
            self.dlm_len = self._io.read_u4le()
            self._debug['dlm_len']['end'] = self._io.pos()
            self._debug['fw_ver']['start'] = self._io.pos()
            self.fw_ver = self._io.read_u2le()
            self._debug['fw_ver']['end'] = self._io.pos()
            self._debug['build_ver']['start'] = self._io.pos()
            self.build_ver = self._io.read_u2le()
            self._debug['build_ver']['end'] = self._io.pos()
            self._debug['extra']['start'] = self._io.pos()
            self.extra = self._io.read_u4le()
            self._debug['extra']['end'] = self._io.pos()
            self._debug['build_time']['start'] = self._io.pos()
            self.build_time = (self._io.read_bytes(16)).decode(u"UTF-8")
            self._debug['build_time']['end'] = self._io.pos()



